-- X-Perl UnitFrames
-- Author: Zek <Boodhoof-EU>
-- License: GNU GPL v3, 29 June 2007 (see LICENSE.txt)

local XPerl_Party_Events = {}
local PartyFrames = {}
local startupDone
local conf, pconf
XPerl_RequestConfig(function(new)
			conf = new
			pconf = new.party
			for k,v in pairs(PartyFrames) do
				v.conf = pconf
			end
		end, "$Revision: 652 $")

local percD = "%d"..PERCENT_SYMBOL

local format = format
local GetNumGroupMembers = GetNumGroupMembers
local UnitIsConnected = UnitIsConnected
local UnitName = UnitName
local UnitIsDead = UnitIsDead
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local XPerl_Unit_GetHealth = XPerl_Unit_GetHealth
local XPerl_Unit_Initialize = XPerl_Unit_Initialize
local partyHeader
local partyAnchor

local XPerl_Party_HighlightCallback

----------------------
-- Loading Function --
----------------------
-- SetFrameArray
local function SetFrameArray(self, value)
	for k,v in pairs(PartyFrames) do
		if (v == self) then
			PartyFrames[k] = nil
			if (XPerl_PartyPetFrames) then
				local petid = "partypet"..strmatch(k, "^party(%d)")
				if (XPerl_PartyPetFrames[petid]) then
					XPerl_PartyPetFrames[petid].unit = nil
					XPerl_PartyPetFrames[petid].ownerid = nil
					XPerl_PartyPetFrames[petid] = nil
				end
			end
		end
	end

	self.unit = value
	if (value) then
		self.targetid = value.."target"
		PartyFrames[value] = self
		if (XPerl_PartyPetFrames and self.petFrame) then
			local petid = "partypet"..strmatch(value, "^party(%d)")
			XPerl_PartyPetFrames[petid] = self.petFrame
			self.petFrame.unit = petid
			self.petFrame.ownerid = value
		end
	end
end
-- onAttrChanged
local function onAttrChanged(self, name, value)
	if (name == "unit") then
		if (value and value ~= "party0") then
			SetFrameArray(self, value)
			if (self.unit ~= value or self.lastName ~= UnitName(value)) then
				if (conf) then
					XPerl_Party_UpdateDisplay(self, true)
				end
				if (XPerl_ArcaneBar_RegisterFrame) then
					XPerl_ArcaneBar_RegisterFrame(self.nameFrame, value)
				end
			end
		else
			SetFrameArray(self)
		end
	end
end
function XPerl_Party_Events_OnLoad(self)
	for i = 1,4 do
		XPerl_BlizzFrameDisable(_G["PartyMemberFrame"..i])
	end
	local events = {"GROUP_ROSTER_UPDATE", "PLAYER_ENTERING_WORLD", "PLAYER_TARGET_CHANGED", "RAID_TARGET_UPDATE"}
	for i,event in pairs(events) do
		self:RegisterEvent(event)
	end

	partyHeader:UnregisterEvent("UNIT_NAME_UPDATE")			-- IMPORTANT! Fix for WoW 2.1 UNIT_NAME_UPDATE lockup issues
	UIParent:UnregisterEvent("GROUP_ROSTER_UPDATE")			-- IMPORTANT! Stops raid framerate lagging when members join/leave/zone

	self:SetScript("OnEvent", XPerl_Party_OnEvent)
	XPerl_RegisterOptionChanger(XPerl_Party_Set_Bits)
	XPerl_Highlight:Register(XPerl_Party_HighlightCallback, self)

	XPerl_Party_Events_OnLoad = nil
end
-- XPerl_Party_OnLoad
function XPerl_Party_OnLoad(self)
	XPerl_SetChildMembers(self)
	self.targetFrame.statsFrame = self.targetFrame.healthBar	-- So the healthbar fades as part of pseudo statsFrame

	partyHeader = XPerl_Party_SecureHeader
	partyAnchor = XPerl_Party_Anchor

	local id = strmatch(self:GetName(), ".+(%d)")
	self:SetID(tonumber(id))
	_G["XPerl_party"..self:GetID()] = self

	if (self:GetID() > 1) then
		self.buffSetup = XPerl_party1.buffSetup
	else
		local BuffOnUpdate, DebuffOnUpdate, BuffUpdateTooltip, DebuffUpdateTooltip
		BuffUpdateTooltip = XPerl_Unit_SetBuffTooltip
		DebuffUpdateTooltip = XPerl_Unit_SetDebuffTooltip

		self.buffSetup = {
			buffScripts = {
				OnEnter = XPerl_Unit_SetBuffTooltip,
				OnUpdate = BuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			debuffScripts = {
				OnEnter = XPerl_Unit_SetDeBuffTooltip,
				OnUpdate = DebuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			updateTooltipBuff = BuffUpdateTooltip,
			updateTooltipDebuff = DebuffUpdateTooltip,
			debuffParent = true,
			debuffSizeMod = 0.4,
			debuffAnchor1 = function(self, b)
				if (pconf.flip) then
					b:SetPoint("RIGHT", self.statsFrame, "LEFT", 0, 0)
				else
					b:SetPoint("LEFT", self.statsFrame, "RIGHT", 0, 0)
				end
			end,
			buffAnchor1 = function(self, b)
				if (pconf.flip) then
					b:SetPoint("TOPRIGHT", self.buffFrame, "TOPRIGHT", 0, 0)
				else
					b:SetPoint("TOPLEFT", self.buffFrame, "TOPLEFT", 0, 0)
				end
			end,
		}
	end

	partyHeader:SetAttribute("child"..self:GetID(), self)
	self.unit = "party"..self:GetID()
	PartyFrames[self.unit] = self

	self.hitIndicator.text:SetPoint("CENTER", self.portraitFrame, "CENTER", 0, 0)
	CombatFeedback_Initialize(self, self.hitIndicator.text, 30)

	local events = {"PARTY_MEMBER_ENABLE", "PARTY_MEMBER_DISABLE", "UNIT_THREAT_LIST_UPDATE", 
					"UNIT_AURA", "UNIT_COMBAT", "UNIT_LEVEL", "UNIT_FACTION", 
					"UNIT_PHASE", "UNIT_OTHER_PARTY_CHANGED", "UNIT_TARGET", 
					"UNIT_FLAGS", "UNIT_DYNAMIC_FLAGS", "PLAYER_FLAGS_CHANGED", 
					"READY_CHECK", "READY_CHECK_CONFIRM", "READY_CHECK_FINISHED"}
	for i,event in pairs(events) do
		self:RegisterUnitEvent(event, self.unit)
	end

	self:SetScript("OnUpdate", XPerl_Party_OnUpdate)
	self:SetScript("OnEvent", XPerl_Party_OnEvent)
	self:SetScript("OnAttributeChanged", onAttrChanged)
	self:SetScript("OnShow", XPerl_Party_UpdateDisplay)
	self.targetFrame:SetScript("OnUpdate", XPerl_Party_Target_OnUpdate)

	XPerl_SecureUnitButton_OnLoad(self, self.unit, nil, _G["PartyMemberFrame"..self:GetID().."DropDown"], XPerl_ShowGenericMenu)
	XPerl_SecureUnitButton_OnLoad(self.nameFrame, self.unit, nil, _G["PartyMemberFrame"..self:GetID().."DropDown"], XPerl_ShowGenericMenu)
	
	self.RangeUpdate, self.targetFrame.time = 0, 0

	XPerl_RegisterHighlight(self.highlight, 3)
	XPerl_RegisterHighlight(self.targetFrame, 3)
	XPerl_RegisterPerlFrames(self, {self.nameFrame, self.statsFrame, self.portraitFrame, self.levelFrame, self.targetFrame})

	self.FlashFrames = {self.nameFrame, self.levelFrame, self.statsFrame, self.portraitFrame}

	if (XPerlDB) then
		self.conf = XPerlDB.party
	end

	if (XPerl_party1 and XPerl_party2 and XPerl_party3 and XPerl_party4) then
		XPerl_Party_OnLoad = nil
	end
end

-- XPerl_Party_HighlightCallback
function XPerl_Party_HighlightCallback(self, updateGUID)
	if not updateGUID then return end

	local f = XPerl_Party_GetUnitFrameByGUID(updateGUID)
	if (f) then
		XPerl_Highlight:SetHighlight(f, updateGUID)
	end
end
-- XPerl_Party_CombatFlash
local function XPerl_Party_CombatFlash(self, elapsed, argNew, argGreen)
	if (XPerl_CombatFlashSet(self, elapsed, argNew, argGreen)) then
		XPerl_CombatFlashSetFrames(self)
	end
end
-- XPerl_Party_SetGray(self)
local function XPerl_Party_SetGray(self)
	XPerl_Unit_SetGrey(self, self.statsFrame.healthBar)	
	XPerl_Unit_HealthBar_Update(self.statsFrame.healthBar, self.unit)
	XPerl_Unit_SetGrey(self, self.statsFrame.manaBar)
	XPerl_Unit_ManaBar_Update(self.statsFrame.manaBar, self.unit)
end
--------------------
-- Buff Functions --
--------------------
-- XPerl_Party_SetDebuffLocation
function XPerl_Party_SetDebuffLocation(self)
	local debuff1 = self.buffFrame.debuff and self.buffFrame.debuff[1]
	if (debuff1) then
		debuff1:ClearAllPoints()

		if (pconf.debuffs.below) then
			local buff1 = self.buffFrame.buff and self.buffFrame.buff[1]
			if (not buff1) then
				if (pconf.flip) then
					debuff1:SetPoint("TOPRIGHT", self.buffFrame, "TOPRIGHT", 0, -20)
				else
					debuff1:SetPoint("TOPLEFT", self.buffFrame, "TOPLEFT", 0, -20)
				end
			else
				if (pconf.flip) then
					debuff1:SetPoint("TOPRIGHT", buff1, "BOTTOMRIGHT", 0, -2)
				else
					debuff1:SetPoint("TOPLEFT", buff1, "BOTTOMLEFT", 0, -2)
				end
			end
		else
			if (self.petFrame and self.petFrame:IsVisible()) then
				if (pconf.flip) then
					debuff1:SetPoint("TOPRIGHT", self.petFrame.nameFrame, "TOPRIGHT", 0, -4)
				else
					debuff1:SetPoint("TOPLEFT", self.petFrame.nameFrame, "TOPRIGHT", 0, -4)
				end
			else
				if (pconf.flip) then
					debuff1:SetPoint("TOPRIGHT", self.statsFrame, "TOPLEFT", 0, -4)
				else
					debuff1:SetPoint("TOPLEFT", self.statsFrame, "TOPRIGHT", 0, -4)
				end
			end

			if (pconf.debuffs.halfSize) then
				local buffWidth = debuff1:GetWidth() * debuff1:GetScale()
				local selfWidth = self:GetWidth() * self:GetScale()
				local maxDebuffWidth = floor(selfWidth / buffWidth)
				local prev
				if (self.perlDebuffs > maxDebuffWidth) then
					self.debuffFrame:SetScale(0.5)

					local Anchor, aHalf, aPrev
					if (pconf.flip) then
						Anchor = "TOPRIGHT"
						aHalf = "BOTTOMRIGHT"
						aPrev = "TOPLEFT"
					else
						Anchor = "TOPLEFT"
						aHalf = "BOTTOMLEFT"
						aPrev = "TOPRIGHT"
					end

					local halfPoint = ceil(self.perlBuffs / 2)
					for k,v in pairs(self.buffFrame.debuff) do
						if (prev) then
							v:ClearAllPoints()
							if (k == halfPoint) then
								v:SetPoint(Anchor, debuff1, aHalf, 0, 0)
							else
								v:SetPoint(Anchor, prev, aPrev, 0, 0)
							end
						end
						prev = v
					end
				else
					self.debuffFrame:SetScale(1)

					local prev
					for k,v in pairs(self.buffFrame.debuff) do
						if (prev) then
							v:ClearAllPoints()
							v:SetPoint("TOPLEFT", prev, "TOPRIGHT", 0, 0)
						end
						prev = v
					end
				end
			else
				self.debuffFrame:SetScale(1)
			end
		end
	end
end
-- XPerl_Party_BuffPositions
local function XPerl_Party_BuffPositions(self)
	if (self.conf) then
		if (pconf.debuffs.below) then
			if (pconf.buffs.wrap) then
				XPerl_Unit_BuffPositions(self, self.buffFrame.buff, self.buffFrame.debuff, self.conf.buffs.size, self.conf.debuffs.size)
			end
		else
			-- Debuffs handled seperately by legacy code, so just do buffs
			if (pconf.buffs.wrap) then
				XPerl_Unit_BuffPositions(self, self.buffFrame.buff, nil, self.conf.buffs.size)
			end
			XPerl_Party_SetDebuffLocation(self)
		end
	end
end
-- XPerl_Party_Buff_UpdateAll
local function XPerl_Party_Buff_UpdateAll(self)
	if (self.conf) then
		if (not pconf.buffs.enable and not pconf.debuffs.enable) then
			self.buffFrame:Hide()
			self.debuffFrame:Hide()
		else
			XPerl_Unit_UpdateBuffs(self, nil, nil, pconf.buffs.castable, pconf.debuffs.curable)
			XPerl_Party_BuffPositions(self)
		end
	
		if (select(2, UnitClass(self.unit)) == "HUNTER") then
			local feigning = UnitIsFeignDeath(self.unit)
			if (feigning ~= self.feigning) then
				self.feigning = feigning
				XPerl_Party_UpdateHealth(self)
			end
		end
	
		XPerl_CheckDebuffs(self, self.unit)
	end
end
-------------------------
-- The Update Function --
-------------------------
-- XPerl_Party_RaidIconUpdate
local function XPerl_Party_RaidIconUpdate(self)
	XPerl_Update_RaidIcon(self.targetFrame.raidIcon, self.unit.."target")
	XPerl_Update_RaidIcon(self.nameFrame.raidIcon, self.unit)
end
-- XPerl_Party_UpdateNotPresentIcon
local function XPerl_Party_UpdateNotPresentIcon(self)
	local inPhase = UnitInPhase(self.unit)
	local inOtherGroup, canInteract = UnitInOtherParty(self.unit)
	if ( inOtherGroup ) then
		self.notPresentIcon.texture:SetTexture("Interface\\PlayerFrame\\whisper-only")
		self.notPresentIcon.texture:SetTexCoord(0.1875, 0.8125, 0.1875, 0.8125)
		self.notPresentIcon.tooltip = PARTY_IN_PUBLIC_GROUP_MESSAGE
		self.notPresentIcon:Show()
	elseif ( not inPhase and UnitExists(self.unit) ) then
		self.notPresentIcon.texture:SetTexture("Interface\\TargetingFrame\\UI-PhasingIcon")
		self.notPresentIcon.texture:SetTexCoord(0.15625, 0.84375, 0.15625, 0.84375)
		self.notPresentIcon.tooltip = PARTY_PHASED_MESSAGE
		self.notPresentIcon:Show()
	else
		self.notPresentIcon:Hide()
	end	
end
-- XPerl_Party_UpdateLevel
local function XPerl_Party_UpdateLevel(self)
	if (self.conf.level) then
		local Level = UnitLevel(self.unit)
		self.levelFrame.text:SetText(Level)		
		if (UnitCanAttack("player", self.unit)) then
			local color = GetQuestDifficultyColor(Level)
			self.levelFrame.text:SetVertexColor(color.r, color.g, color.b)
		else				
			self.levelFrame.text:SetVertexColor(1.0, 0.82, 0.0)
		end
	end
end
-- XPerl_Party_UpdateClass
local function XPerl_Party_UpdateClass(self)
	if (UnitIsPlayer(self.unit)) then
		local l, r, t, b = XPerl_ClassPos(select(2, UnitClass(self.unit)))
		self.levelFrame.classTexture:SetTexCoord(l, r, t, b)
	end

	if (pconf.classIcon) then
		self.levelFrame.classTexture:Show()
	else
		self.levelFrame.classTexture:Hide()
	end
end
-- XPerl_Party_UpdateCombat
local function XPerl_Party_UpdateCombat(self)
	if (UnitIsVisible(self.unit) and UnitAffectingCombat(self.unit)) then
		self.nameFrame.combatIcon:Show()
	else
		self.nameFrame.combatIcon:Hide()
	end
end
-- XPerl_Party_UpdateLeader
local function XPerl_Party_UpdateLeader(self)
	-- Leader
	if (UnitIsGroupLeader(self.unit)) then
		self.nameFrame.leaderIcon:Show()
	else
		self.nameFrame.leaderIcon:Hide()
	end
	-- Loot Master
	local lootMethod, lootMaster
	lootMethod, lootMaster = GetLootMethod()
	if (self.unit == "party"..(lootMaster or "nil")) then
		self.nameFrame.masterIcon:Show()
	else
		self.nameFrame.masterIcon:Hide()
	end	
end
-- XPerl_Party_UpdatePVP
local function XPerl_Party_UpdatePVP(self)
	local factionGroup = UnitFactionGroup(self.unit)
	if (self.conf.pvpIcon and UnitIsPVPFreeForAll(self.unit)) then
		self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-FFA")
		self.nameFrame.pvpIcon:Show()
	elseif (self.conf.pvpIcon and factionGroup and factionGroup ~= "Neutral" and UnitIsPVP(self.unit)) then
		self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..factionGroup)
		self.nameFrame.pvpIcon:Show()
	else
		self.nameFrame.pvpIcon:Hide()
	end

	if (UnitIsVisible(self.unit) and UnitIsCharmed(self.unit)) then
		self.nameFrame.warningIcon:Show()
	else
		self.nameFrame.warningIcon:Hide()
	end
end
-- XPerl_Party_UpdateAssignedRoles
local function XPerl_Party_UpdateAssignedRoles(self)	
	local unit = self.unit
	local icon = self.nameFrame.roleIcon
	--if (select(2, IsInInstance()) == "party") then
		role = UnitGroupRolesAssigned(unit)
		if role == "TANK" then
			icon:SetTexture("Interface\\GroupFrame\\UI-Group-MainTankIcon")
		elseif role == "HEALER" then
			icon:SetTexture("Interface\\Addons\\XPerl\\Images\\XPerl_RoleHealer_old")
		elseif role == "DAMAGER" then
			icon:SetTexture("Interface\\GroupFrame\\UI-Group-MainAssistIcon")
		end
	--end
	icon:ClearAllPoints()
	if (self.nameFrame.masterIcon:IsShown()) then
		icon:SetPoint("LEFT", self.nameFrame.masterIcon, "RIGHT")
	elseif (self.nameFrame.leaderIcon:IsShown()) then
		icon:SetPoint("LEFT", self.nameFrame.leaderIcon, "RIGHT")
	else
		icon:SetPoint("TOPLEFT", 10, 5)
	end
	
	icon:SetTexCoord(0,1,0,1)
	if ( role == "TANK" or role == "HEALER" or role == "DAMAGER") then
		if not (conf and conf.xperlOldroleicons) then
			icon:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES")
	       	icon:SetTexCoord(GetTexCoordsForRoleSmallCircle(role))
		end
		icon:Show()
	else
        icon:Hide()
	end
end
-- XPerl_Party_UpdateAllAssignedRoles
local function XPerl_Party_UpdateAllAssignedRoles()
	for unit,frame in pairs(PartyFrames) do
		if (frame:IsShown()) then
			XPerl_Party_UpdateAssignedRoles(frame)
		end
	end
end

-- XPerl_Party_UpdateRange
local function XPerl_Party_UpdateRange(self, overrideUnit)
	local unit = overrideUnit or self.unit
	if (unit) then
		if (not pconf.range30yard or CheckInteractDistance(self.unit, 1) or not UnitIsConnected(self.unit)) then
			self.nameFrame.rangeIcon:Hide()
		else
			self.nameFrame.rangeIcon:Show()
			self.nameFrame.rangeIcon:SetAlpha(1)
		end
		--if (UnitInVehicle(self.unit) and pconf.range30yard) then--Not sure if this is proper way to do it, so this pretty much forces anyone in a vehicle to show out of range.
		--	self.nameFrame.rangeIcon:Show()
		--	self.nameFrame.rangeIcon:SetAlpha(1)
		--end
	end
end

-- XPerl_Party_SingleGroup
function XPerl_Party_SingleGroup()
	local num = GetNumGroupMembers()
	if (num > 5) then
		return false
	end
	for i = 1,num do
		local name, rank, group = GetRaidRosterInfo(i)
		if (group > 1) then
			return false
		end
	end
	return true
end
-- CheckRaid
local function CheckRaid()
	if (InCombatLockdown()) then
		XPerl_OutOfCombatQueue[CheckRaid] = false
	else
		partyAnchor:StopMovingOrSizing()

		local singleGroup = XPerl_Party_SingleGroup()
		if (not pconf or (pconf.inRaid or (pconf.smallRaid and singleGroup) or GetNumGroupMembers() == 0)) then
			if (not partyHeader:IsShown()) then
				partyHeader:Show()
			end
		else
			if (partyHeader:IsShown()) then
				partyHeader:Hide()
			end
		end
	end
end

-- XPerl_Party_TargetUpdateHealth
local function XPerl_Party_TargetUpdateHealth(self)
	local tf = self.targetFrame
	local hp, hpMax, percent = XPerl_Unit_GetHealth(self.targetid)
	if (pconf.target.large and hp ~= tf.lastHP or hpMax ~= tf.lastHPMax or GetTime() > tf.lastUpdate + 5000) then	
		tf.lastHP, tf.lastHPMax = hp, hpMax
		tf.lastUpdate = GetTime()

		tf.healthBar:SetMinMaxValues(0, hpMax)
		tf.healthBar:SetValue(hp)
		tf.healthBar.text:SetFormattedText(percD, percent)
		tf.healthBar.text:Show()

		if (UnitIsDeadOrGhost(self.targetid)) then
			tf.healthBar:SetStatusBarColor(0.5, 0.5, 0.5, 1)
			tf.healthBar.bg:SetVertexColor(0.5, 0.5, 0.5, 0.5)
			if (UnitIsDead(self.targetid)) then
				tf.healthBar.text:SetText(XPERL_LOC_DEAD)
			else
				tf.healthBar.text:SetText(XPERL_LOC_GHOST)
			end
		else
			XPerl_Unit_SetSmoothBarColor(self.targetFrame.healthBar, percent)
		end

		if (UnitAffectingCombat(self.targetid)) then
			tf.combatIcon:SetTexCoord(0.5, 1.0, 0.0, 0.5)
			tf.combatIcon:Show()
		else
			tf.combatIcon:Hide()
		end

		if (pconf.pvpIcon and UnitIsPVP(self.targetid)) then
			tf.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..(UnitFactionGroup(self.targetid) or "FFA"))
			tf.pvpIcon:Show()
		else
			tf.pvpIcon:Hide()
		end
	end
end
-- XPerl_Party_UpdateTarget
local function XPerl_Party_UpdateTarget(self)
	if (pconf.target.enable) then
		if (self.targetid and UnitIsConnected(self.unit) and UnitExists(self.unit) and UnitIsVisible(self.unit)) then
			local targetname = UnitName(self.targetid)
			if (targetname and targetname ~= UNKNOWNOBJECT) then
				self.targetFrame.text:SetText(targetname)
				XPerl_SetUnitNameColor(self.targetFrame.text, self.targetid)
				XPerl_Party_TargetUpdateHealth(self)
				XPerl_Party_RaidIconUpdate(self)
			end
		end
	end
end

-- XPerl_Party_UpdateDisplay
function XPerl_Party_UpdateDisplay(self, less)
	if (self.conf and self.unit and UnitExists(self.unit)) then
		if (not less) then
			XPerl_Party_UpdateLevel(self)
		end

		XPerl_Unit_UpdateName(self)
		XPerl_Unit_UpdatePortrait(self)
		XPerl_Party_SetGray(self)
		XPerl_Party_RaidIconUpdate(self)
		XPerl_Party_UpdateAssignedRoles(self)
		XPerl_Party_UpdateNotPresentIcon(self)
		XPerl_Party_UpdateClass(self)
		XPerl_Party_UpdateLeader(self)
		XPerl_Party_UpdateCombat(self)
		XPerl_Party_UpdatePVP(self)
		XPerl_Unit_UpdateReadyState(self)
		XPerl_Party_Buff_UpdateAll(self)
		XPerl_Party_UpdateTarget(self)
		XPerl_UpdateSpellRange(self, self.unit)
	end
end
-- XPerl_Party_UpdateDisplayAll
function XPerl_Party_UpdateDisplayAll()
	for i,frame in pairs(PartyFrames) do
		if (frame.unit) then
			XPerl_Party_UpdateDisplay(frame)
		end
	end
end

-- XPerl_Party_OnUpdate
function XPerl_Party_OnUpdate(self, elapsed, ...)
	if (not self.unit) then
		return
	end	

	if (self.PlayerFlash) then
		XPerl_Party_CombatFlash(self, elapsed, false)
	end

	self.RangeUpdate = self.RangeUpdate + elapsed
	if (self.RangeUpdate >= 0.2) then
		self.RangeUpdate = 0
		local unit = self.unit

		XPerl_Party_UpdateRange(self, unit)
		XPerl_UpdateSpellRange(self, unit)
		XPerl_UpdateSpellRange(self.targetFrame, self.targetid)
	end
	
	CombatFeedback_OnUpdate(self, elapsed)
end
-- XPerl_Party_Target_OnUpdate
function XPerl_Party_Target_OnUpdate(self, elapsed)
	self.time = elapsed + (self.time or 0)
	if (self.time >= 0.5) then
		self.time = 0
		XPerl_Party_UpdateTarget(self:GetParent())
	end
end

-------------------
-- Event Handler --
-------------------
function XPerl_Party_OnEvent(self, event, unitID, ...)
	--print (self:GetName().."-"..event.."-"..(unitID or "nil"))
	local func = XPerl_Party_Events[event]
	if (func) then
		if (strfind(event, "^UNIT_") and event ~= "UNIT_THREAT_LIST_UPDATE") then
			local f = PartyFrames[unitID]
			if (f) then
				func(f, ...)
			end
		else
			func(self, unitID, ...)
		end
	else
		XPerl_Unit_OnEvent(self, event, unitID)
	end
end

-- PLAYER_ENTERING_WORLD
function XPerl_Party_Events:PLAYER_ENTERING_WORLD()
	UIParent:UnregisterEvent("GROUP_ROSTER_UPDATE")	-- Re-do, in case
	if (not startupDone) then
		startupDone = true
		XPerl_ProtectedCall(XPerl_Party_SetInitialAttributes)
		CheckRaid()
	end
	XPerl_Party_UpdateDisplayAll()
end
-- UNIT_LEVEL
function XPerl_Party_Events:UNIT_LEVEL()
	XPerl_Party_UpdateLevel(self)
end
-- RAID_TARGET_UPDATE
function XPerl_Party_Events:RAID_TARGET_UPDATE()
	for i,frame in pairs(PartyFrames) do
		if (frame.unit) then
			XPerl_Party_RaidIconUpdate(frame)
		end
	end
end
-- READY_CHECK
function XPerl_Party_Events:READY_CHECK(a, b, c)
	if (self.unit) then
		XPerl_Unit_UpdateReadyState(self)
	end
end
XPerl_Party_Events.READY_CHECK_CONFIRM = XPerl_Party_Events.READY_CHECK
XPerl_Party_Events.READY_CHECK_FINISHED = XPerl_Party_Events.READY_CHECK
-- PLAYER_FLAGS_CHANGED
function XPerl_Party_Events:PLAYER_FLAGS_CHANGED(unit)
	XPerl_Party_SetGray(self)
end
-- UNIT_FACTION
function XPerl_Party_Events:UNIT_FACTION(unit)
	XPerl_Unit_UpdateName(self)
	XPerl_Party_SetGray(self)
	XPerl_Party_UpdateCombat(self)
	XPerl_Party_UpdatePVP(self)
	XPerl_Unit_ThreatStatus(self)
end
XPerl_Party_Events.UNIT_FLAGS = XPerl_Party_Events.UNIT_FACTION
XPerl_Party_Events.UNIT_DYNAMIC_FLAGS = XPerl_Party_Events.UNIT_FACTION
-- PARTY_MEMBER_ENABLE
function XPerl_Party_Events:PARTY_MEMBER_ENABLE()
	XPerl_Party_UpdateDisplay(self)
end
XPerl_Party_Events.PARTY_MEMBER_DISABLE = XPerl_Party_Events.PARTY_MEMBER_ENABLE
XPerl_Party_Events.UNIT_OTHER_PARTY_CHANGED = XPerl_Party_Events.PARTY_MEMBER_ENABLE
XPerl_Party_Events.UNIT_PHASE = XPerl_Party_Events.PARTY_MEMBER_ENABLE
-- UNIT_COMBAT
function XPerl_Party_Events:UNIT_COMBAT(...)
	local action, descriptor, damage, damageType = select(1, ...)
	
	if (pconf.hitIndicator and pconf.portrait) then
		CombatFeedback_OnCombatEvent(self, action, descriptor, damage, damageType)
	end

	XPerl_Party_UpdateCombat(self)
	if (action == "HEAL") then
		XPerl_Party_CombatFlash(self, 0, true, true)
	elseif (damage and damage > 0) then
		XPerl_Party_CombatFlash(self, 0, true)
	end
end
-- UNIT_AURA
function XPerl_Party_Events:UNIT_AURA()
	XPerl_Party_Buff_UpdateAll(self)
end

local function updatePartyThreat(immediate)
	for unitid,frame in pairs(PartyFrames) do
		if (frame:IsShown()) then
			XPerl_Unit_ThreatStatus(frame, nil, immediate)
		end
	end
end
function XPerl_Party_Events:UNIT_THREAT_LIST_UPDATE(unit)
	if (unit == "target") then
		updatePartyThreat()
	end
end
function XPerl_Party_Events:PLAYER_TARGET_CHANGED()
	updatePartyThreat(true)
end
-- UNIT_TARGET
function XPerl_Party_Events:UNIT_TARGET()
	XPerl_Party_UpdateTarget(self)
	updatePartyThreat(true)
end


-- XPerl_Party_GetUnitFrameByUnit
function XPerl_Party_GetUnitFrameByUnit(unitid)
	return PartyFrames[unitid]
end

do
	local rosterGuids
	-- XPerl_Party_GetUnitFrameByGUID
	function XPerl_Party_GetUnitFrameByGUID(guid)
		local unitid = rosterGuids and rosterGuids[guid]
		if (unitid) then
			return PartyFrames[unitid]
		end
	end

	local function BuildGuidMap()
		if (GetNumSubgroupMembers() > 0) then
			rosterGuids = XPerl_GetReusableTable()
			for i = 1,GetNumSubgroupMembers() do
				local guid = UnitGUID("party"..i)
				if (guid) then
					rosterGuids[guid] = "party"..i
				end
			end
		else
			rosterGuids = XPerl_FreeTable(rosterGuids)
		end
	end
	
	-- GROUP_ROSTER_UPDATE
	function XPerl_Party_Events:GROUP_ROSTER_UPDATE()
		BuildGuidMap()
		CheckRaid()
		XPerl_SetHighlights()
		XPerl_Party_UpdateDisplayAll()
	end
end

---- Moving stuff ----
-- XPerl_Party_GetGap
function XPerl_Party_GetGap()
	--return floor(floor((XPerl_party1:GetBottom() - XPerl_party2:GetTop() + 0.01) * 100) / 100)
	return pconf.spacing
end

-- XPerl_Party_SetGap
function XPerl_Party_SetGap(newGap)
	if (type(newGap) ~= "number") then
		return
	end
	pconf.spacing = newGap
	XPerl_Party_SetMainAttributes()
	XPerl_Party_Virtual(true)
end

-- CalcWidth
local function CalcWidth(self)
	--return XPerl_party1Highlight:GetWidth()

	local w = self.statsFrame:GetWidth() or 0

	if (pconf and pconf.portrait) then
		w = w + (self.portraitFrame:GetWidth() or 0) - 2

		if (pconf.level or pconf.classIcon) then
			w = w + (self.levelFrame:GetWidth() or 0) - 2
		end
	else
		w = w + (self.levelFrame:GetWidth() or 0) - 2
	end

	return w
end
-- CalcHeight
local function CalcHeight(self)
	--return XPerl_party1Highlight:GetHeight()

	if (pconf and pconf.portrait) then
		return self.portraitFrame:GetHeight() + 1
	end

	local h = self.statsFrame:GetHeight()

	if (pconf and pconf.name) then
		h = h + self.nameFrame:GetHeight() - 2
	end

	return h
end

-- XPerl_Party_SetWidth
function XPerl_Party_SetWidth(self)

	pconf.size.width = max(0, pconf.size.width or 0)

	local width = (36 * (pconf.percent or 0)) + 106	-- 136 enabled, 106 disabled
	self.statsFrame:SetWidth(width + pconf.size.width)
	self:SetWidth(CalcWidth(self))

	self.nameFrame:SetWidth(106 + (pconf.size.width / 2))

	XPerl_StatsFrameSetup(self)
end

-- XPerl_Party_Set_Bits
function XPerl_Party_Set_Bits1(self)
	if (InCombatLockdown()) then
		XPerl_OutOfCombatQueue[XPerl_Party_Set_Bits1] = self
		return
	end
	XPerl_Unit_Initialize(self, self.nameFrame.text, self.statsFrame.healthBar, self.statsFrame.manaBar, self.statsFrame.expectedHealth)
	
	self.portraitFrame:ClearAllPoints()
	self.nameFrame:ClearAllPoints()
	self.levelFrame:ClearAllPoints()
	self.statsFrame:ClearAllPoints()
	self.levelFrame.classTexture:ClearAllPoints()
	self.levelFrame.text:ClearAllPoints()
	self.highlight:ClearAllPoints()

	if (not pconf.portrait) then
		self.portraitFrame:Hide()

		self.levelFrame:SetWidth(30)
		self.levelFrame:SetHeight(41)

		if (pconf.flip) then
			self.nameFrame:SetPoint("TOPRIGHT", self, "TOPRIGHT", 0, 0)
			self.levelFrame:SetPoint("TOPRIGHT", self.nameFrame, "BOTTOMRIGHT", 0, 3)
			self.statsFrame:SetPoint("TOPRIGHT", self.levelFrame, "TOPLEFT", 2, 0)

			self.levelFrame.text:SetPoint("BOTTOM", self.levelFrame, "BOTTOM", 0, 4)
			self.levelFrame.classTexture:SetPoint("TOPRIGHT", self.levelFrame, "TOPRIGHT", -5, -5)

			self.buffFrame:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT", -2, 0)

			self.highlight:SetPoint("TOPRIGHT", self.nameFrame, "TOPRIGHT", 0, 0)
			self.highlight:SetPoint("BOTTOMLEFT", self.statsFrame, "BOTTOMLEFT", 0, 0)
		else
			self.nameFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
			self.levelFrame:SetPoint("TOPLEFT", self.nameFrame, "BOTTOMLEFT", 0, 3)
			self.statsFrame:SetPoint("TOPLEFT", self.levelFrame, "TOPRIGHT", -2, 0)

			self.levelFrame.text:SetPoint("BOTTOM", self.levelFrame, "BOTTOM", 0, 4)
			self.levelFrame.classTexture:SetPoint("TOPLEFT", self.levelFrame, "TOPLEFT", 5, -5)

			self.buffFrame:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 2, 0)

			self.highlight:SetPoint("TOPLEFT", self.nameFrame, "TOPLEFT", 0, 0)
			self.highlight:SetPoint("BOTTOMRIGHT", self.statsFrame, "BOTTOMRIGHT", 0, 0)
		end
	else
		self.portraitFrame:Show()

		self.levelFrame:SetWidth(27)
		self.levelFrame:SetHeight(22)

		if (pconf.flip) then
			self.levelFrame:SetPoint("TOPRIGHT", self, "TOPRIGHT", 0, 0)
			self.portraitFrame:SetPoint("TOPRIGHT", self.levelFrame, "TOPLEFT", 2, 0)
			self.nameFrame:SetPoint("TOPRIGHT", self.portraitFrame, "TOPLEFT", 2, 0)
			self.statsFrame:SetPoint("TOPRIGHT", self.nameFrame, "BOTTOMRIGHT", 0, 3)

			self.levelFrame.text:SetPoint("CENTER", 0, 0)
			self.levelFrame.classTexture:SetPoint("BOTTOMLEFT", self.portraitFrame, "BOTTOMRIGHT", 0, 3)

			self.buffFrame:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT", -2, 0)

			self.highlight:SetPoint("TOPRIGHT", self.levelFrame, "TOPRIGHT", 0, 0)
			self.highlight:SetPoint("BOTTOMLEFT", self.statsFrame, "BOTTOMLEFT", 0, 0)
		else
			self.levelFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
			self.portraitFrame:SetPoint("TOPLEFT", self.levelFrame, "TOPRIGHT", -2, 0)
			self.nameFrame:SetPoint("TOPLEFT", self.portraitFrame, "TOPRIGHT", -2, 0)
			self.statsFrame:SetPoint("TOPLEFT", self.nameFrame, "BOTTOMLEFT", 0, 3)

			self.levelFrame.text:SetPoint("CENTER", 0, 0)
			self.levelFrame.classTexture:SetPoint("BOTTOMRIGHT", self.portraitFrame, "BOTTOMLEFT", 0, 3)

			self.buffFrame:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 2, 0)

			self.highlight:SetPoint("TOPLEFT", self.levelFrame, "TOPLEFT", 0, 0)
			self.highlight:SetPoint("BOTTOMRIGHT", self.statsFrame, "BOTTOMRIGHT", 0, 0)
		end
	end

	if (pconf.level) then
		self.levelFrame.text:Show()
		self.levelFrame:Show()

		if (pconf.portrait) then
			self.levelFrame:SetWidth(27)
		else
			self.levelFrame:SetWidth(30)
		end
	else
		self.levelFrame.text:Hide()
	end

	if (pconf.classIcon) then
		self.levelFrame.classTexture:Show()
		self.levelFrame:Show()

		if (pconf.portrait) then
			self.levelFrame:SetWidth(27)
		else
			self.levelFrame:SetWidth(30)
		end
	else
		self.levelFrame.classTexture:Hide()

		if (not pconf.level) then
			self.levelFrame:SetWidth(2)
			self.levelFrame:Hide()
		end
	end

	if (pconf.percent) then
		self.statsFrame.healthBar.percent:Show()
		self.statsFrame.manaBar.percent:Show()
	else
		self.statsFrame.healthBar.percent:Hide()
		self.statsFrame.manaBar.percent:Hide()
	end

	local height = ((pconf.name or 0) * 22) + 2		-- 24 when enabled, 2 when disabled

	self.targetFrame:ClearAllPoints()
	self.nameFrame:SetHeight(height)

	if (pconf.name) then
		self.nameFrame:Show()
		if (pconf.flip) then
			self.targetFrame:SetPoint("BOTTOMRIGHT", self.nameFrame, "BOTTOMLEFT", 2, 0)
		else
			self.targetFrame:SetPoint("BOTTOMLEFT", self.nameFrame, "BOTTOMRIGHT", -2, 0)
		end
	else
		self.nameFrame:Hide()
		if (pconf.flip) then
			self.targetFrame:SetPoint("TOPRIGHT", self.statsFrame, "BOTTOMRIGHT", 2, 0)
		else
			self.targetFrame:SetPoint("TOPLEFT", self.statsFrame, "BOTTOMLEFT", -2, 0)
		end
	end

	if (pconf.target.large) then
		self.targetFrame.healthBar.text:SetTextColor(1, 1, 1)
		self.targetFrame:SetHeight(28)
		self.targetFrame.healthBar:Show()
	else
		self.targetFrame:SetHeight(20)
		self.targetFrame.healthBar:Hide()
	end

	self.targetFrame:SetWidth(pconf.target.size)

	pconf.buffs.size = tonumber(pconf.buffs.size) or 20
	XPerl_SetBuffSize(self)

	local function SetAllBuffs(self, buffs, debuff)
		local prevAnchor
		if (pconf.flip) then
			prevAnchor = "TOPRIGHT"
		else
			prevAnchor = "TOPLEFT"
		end
		if (buffs) then
			local prev = self
			for k,v in pairs(buffs) do
				v:ClearAllPoints()
				if (pconf.flip) then
					v:SetPoint("TOPRIGHT", prev, prevAnchor, -(1 + debuff), 0)
				else
					v:SetPoint("TOPLEFT", prev, prevAnchor, 1 + debuff, 0)
				end
				prev = v
				if (pconf.flip) then
					prevAnchor = "TOPLEFT"
				else
					prevAnchor = "TOPRIGHT"
				end
			end
		end
	end

	if (not pconf.debuffs.halfSize) then
		self.debuffFrame:SetScale(1)
	end
	SetAllBuffs(self.debuffFrame, self.buffFrame.debuff, 1)
	SetAllBuffs(self.buffFrame, self.buffFrame.buff, 0)

	self.buffOptMix = nil
	self.debuffFrame:SetScale(1)
	XPerl_Party_BuffPositions(self)

	XPerl_Party_SetWidth(self)

	if (pconf.target.enable) then
		RegisterUnitWatch(self.targetFrame)
	else
		UnregisterUnitWatch(self.targetFrame)
		self.targetFrame:Hide()
	end

	if (self:IsShown()) then
		XPerl_Party_UpdateDisplay(self)
	end

	self.petFrame = _G["XPerl_partypet"..self:GetID()]
	if (self.petFrame) then
		self.petFrame:ClearAllPoints()
		if (pconf.flip) then
			self.petFrame:SetPoint("TOPRIGHT", self.statsFrame, "TOPLEFT", 2, 0)
		else
			self.petFrame:SetPoint("TOPLEFT", self.statsFrame, "TOPRIGHT", -2, 0)
		end
	end

	if (XPerl_Voice) then
		XPerl_Voice:Register(self, true)
	end
end

-- XPerl_Party_Set_Bits
function XPerl_Party_Set_Bits()
	if (InCombatLockdown()) then
		XPerl_OutOfCombatQueue[XPerl_Party_Set_Bits] = false
		return
	end

	partyAnchor:SetScale(pconf.scale)
	XPerl_SavePosition(partyAnchor, true)

	if (XPerlDB) then
		conf = XPerlDB
		pconf = XPerlDB.party
		for k,v in pairs(PartyFrames) do
			v.conf = pconf
			XPerl_Party_Set_Bits1(v)
		end
	end

	XPerl_ProtectedCall(XPerl_Party_SetInitialAttributes())

	if (XPerl_Party_AnchorVirtual:IsShown()) then
		XPerl_Party_Virtual(true)
	end
	XPerl_Party_UpdateAllAssignedRoles()
end

-- XPerl_Party_SetInitialAttributes()
function XPerl_Party_SetInitialAttributes()
	partyHeader:Hide()
	XPerl_Party_SetMainAttributes()
	CheckRaid()
end

-- XPerl_Party_SetMainAttributes
function XPerl_Party_SetMainAttributes()
	partyAnchor:StopMovingOrSizing()

	partyHeader:ClearAllPoints()
	if (pconf.anchor == "TOP") then
		partyHeader:SetPoint("TOPLEFT", partyAnchor, "TOPLEFT", 0, 0)
		partyHeader:SetAttribute("xOffset", 0)
		partyHeader:SetAttribute("yOffset", -pconf.spacing)
	elseif (pconf.anchor == "LEFT") then
		partyHeader:SetPoint("TOPLEFT", partyAnchor, "TOPLEFT", 0, 0)
		partyHeader:SetAttribute("xOffset", pconf.spacing)
		partyHeader:SetAttribute("yOffset", 0)
	elseif (pconf.anchor == "BOTTOM") then
		partyHeader:SetPoint("BOTTOMLEFT", partyAnchor, "BOTTOMLEFT", 0, 0)
		partyHeader:SetAttribute("xOffset", 0)
		partyHeader:SetAttribute("yOffset", pconf.spacing)
	elseif (pconf.anchor == "RIGHT") then
		partyHeader:SetPoint("BOTTOMRIGHT", partyAnchor, "BOTTOMRIGHT", 0, 0)
		partyHeader:SetAttribute("xOffset", -pconf.spacing)
		partyHeader:SetAttribute("yOffset", 0)
	end

	partyHeader:SetAttribute("point", pconf.anchor or "TOP")
end

-- XPerl_Party_Virtual
function XPerl_Party_Virtual(on)
	local virtual = XPerl_Party_AnchorVirtual
	if (on) then
		local w = CalcWidth(XPerl_party1)
		local h = CalcHeight(XPerl_party1)

		virtual:ClearAllPoints()
		if (pconf.anchor == "TOP") then
			virtual:SetPoint("TOPLEFT", partyAnchor, "TOPLEFT", 0, 0)
			virtual:SetHeight((h * 4) + (pconf.spacing * 3))
			virtual:SetWidth(w)

		elseif (pconf.anchor == "LEFT") then
			virtual:SetPoint("TOPLEFT", partyAnchor, "TOPLEFT", 0, 0)
			virtual:SetHeight(h)
			virtual:SetWidth(w * 4 + (pconf.spacing * 3))

		elseif (pconf.anchor == "BOTTOM") then
			virtual:SetPoint("BOTTOMLEFT", partyAnchor, "BOTTOMLEFT", 0, 0)
			virtual:SetHeight((h * 4) + (pconf.spacing * 3))
			virtual:SetWidth(w)

		elseif (pconf.anchor == "RIGHT") then
			virtual:SetPoint("TOPRIGHT", partyAnchor, "TOPRIGHT", 0, 0)
			virtual:SetHeight(h)
			virtual:SetWidth(w * 4 + (pconf.spacing * 3))
		end

		virtual:SetBackdropColor(0, 0, 0, 1)
		virtual:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
		virtual:Lower()
		virtual:Show()
	else
		virtual:Hide()
	end
end
