-- X-Perl UnitFrames
-- Author: Zek <Boodhoof-EU>
-- License: GNU GPL v3, 29 June 2007 (see LICENSE.txt)

local XPerl_Party_Pet_Events = {}
local conf, pconf, petconf
local PartyPetFrames = {}
XPerl_RequestConfig(function(New) conf = New
			pconf = New.party
			petconf = New.partypet
			for k,v in pairs(PartyPetFrames) do
				v.conf = pconf
			end
		end, "$Revision: 652 $")

local new, del, copy = XPerl_GetReusableTable, XPerl_FreeTable, XPerl_CopyTable

local GetNumGroupMembers = GetNumGroupMembers
local XPerl_Unit_Initialize = XPerl_Unit_Initialize
local XPerl_Player_Pet_HighlightCallback

----------------------
-- Loading Function --
----------------------
function XPerl_Party_Pet_OnLoadEvents(self)
	self.time = 0

	self:RegisterEvent("GROUP_ROSTER_UPDATE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")

	-- Set here to reduce amount of function calls made
	self:SetScript("OnEvent", XPerl_Party_Pet_OnEvent)

	XPerl_RegisterOptionChanger(XPerl_Party_Pet_Set_Bits, "PartyPet")
	XPerl_Highlight:Register(XPerl_Party_Pet_HighlightCallback, self)

	XPerl_Party_Pet_OnLoadEvents = nil
end

local XPerl_Party_Pet_UpdateGUIDs
do
	local guids
	-- XPerl_Party_Pet_UpdateGUIDs
	function XPerl_Party_Pet_UpdateGUIDs()
		del(guids)
		guids = new()
		for i = 1,GetNumSubgroupMembers() do
			local id = "partypet"..i
			if (UnitExists(id)) then
				guids[UnitGUID(id)] = PartyPetFrames[id]
			end
		end
	end

	-- XPerl_Party_Pet_GetUnitFrameByGUID
	function XPerl_Party_Pet_GetUnitFrameByGUID(guid)
		return guids and guids[guid]
	end

	-- XPerl_Party_Pet_HighlightCallback
	function XPerl_Party_Pet_HighlightCallback(self, updateGUID)
		local f = guids and guids[updateGUID]
		if (f) then
			XPerl_Highlight:SetHighlight(f, updateGUID)
		end
	end
end

-- XPerl_Party_Pet_GetUnitFrameByUnit
function XPerl_Party_Pet_GetUnitFrameByUnit(unitid)
	return PartyPetFrames[unitid]
end

-- XPerl_Party_Pet_OnLoad
function XPerl_Party_Pet_OnLoad(self)
	XPerl_SetChildMembers(self)

	local id = strmatch(self:GetName(), ".+(%d)")
	self:SetID(tonumber(id))
	_G["XPerl_partypet"..self:GetID()] = self

	if (self:GetID() > 1) then
		self.buffSetup = XPerl_partypet1.buffSetup
	else
		local BuffOnUpdate, DebuffOnUpdate, BuffUpdateTooltip, DebuffUpdateTooltip
		BuffUpdateTooltip = XPerl_Unit_SetBuffTooltip
		DebuffUpdateTooltip = XPerl_Unit_SetDeBuffTooltip
		
		self.buffSetup = {
			buffScripts = {
				OnEnter = XPerl_Unit_SetBuffTooltip,
				OnUpdate = BuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			debuffScripts = {
				OnEnter = XPerl_Unit_SetDeBuffTooltip,
				OnUpdate = DebuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			updateTooltipBuff = BuffUpdateTooltip,
			updateTooltipDebuff = DebuffUpdateTooltip,
			debuffParent = true,
			debuffSizeMod = 0,
			debuffAnchor1 = function(self, b)
						local relation = self.buffFrame.buff and self.buffFrame.buff[1]
						if (not relation) then
							relation = XPerl_GetBuffButton(self, 1, 0, 1)
						end
						if (relation) then
							if (pconf.flip) then
								b:SetPoint("TOPRIGHT", relation, "BOTTOMRIGHT", 0, 0)
							else
								b:SetPoint("TOPLEFT", relation, "BOTTOMLEFT", 0, 0)
							end
						else
							if (pconf.flip) then
								b:SetPoint("TOPRIGHT", 0, -14)
							else
								b:SetPoint("TOPLEFT", 0, -14)
							end
						end
					end,
			buffAnchor1 =	function(self, b)
						if (pconf.flip) then
							b:SetPoint("TOPRIGHT", 0, 0)
						else
							b:SetPoint("TOPLEFT", 0, 0)
						end
					end,
		}
	end
		
	self.isPet = true
	self.unit = "partypet"..self:GetID()
	PartyPetFrames[self.unit] = self

	local events = {"UNIT_COMBAT", "UNIT_LEVEL", "UNIT_FACTION", "UNIT_AURA", "UNIT_DYNAMIC_FLAGS", "UNIT_FLAGS"}
	for i,event in pairs(events) do
		self:RegisterUnitEvent(event, self.unit)
	end
	
	-- Set here to reduce amount of function calls made
	self:SetScript("OnEvent", XPerl_Party_Pet_OnEvent)
	self:SetScript("OnUpdate", XPerl_Party_Pet_OnUpdate)
	self:SetScript("OnShow", XPerl_Party_Pet_UpdateDisplay)
	
	XPerl_SecureUnitButton_OnLoad(self, nil, nil, nil, XPerl_ShowGenericMenu)
	XPerl_SecureUnitButton_OnLoad(self.nameFrame, nil, nil, nil, XPerl_ShowGenericMenu)
	
	self:SetAttribute("useparent-unit", true)
	self:SetAttribute("unitsuffix", "pet")
	self.nameFrame:SetAttribute("useparent-unit", true)
	self.nameFrame:SetAttribute("unitsuffix", "pet")

	self.time = 0
	
	XPerl_RegisterHighlight(self.highlight, 2)
	XPerl_RegisterPerlFrames(self, {self.nameFrame, self.statsFrame})

	self.FlashFrames = {self.nameFrame, self.statsFrame}

	if (XPerlDB) then
		self.conf = conf.partypet
	end
end

-- XPerl_Party_Pet_CombatFlash
local function XPerl_Party_Pet_CombatFlash(self, elapsed, argNew, argGreen)
	if (XPerl_CombatFlashSet (self, elapsed, argNew, argGreen)) then
		XPerl_CombatFlashSetFrames(self)
	end
end
-- XPerl_Party_Pet_SetGray
local function XPerl_Party_Pet_SetGray(self)
	XPerl_Unit_SetGrey(self, self.statsFrame.healthBar)	
	XPerl_Unit_HealthBar_Update(self.statsFrame.healthBar, self.unit)
	XPerl_Unit_SetGrey(self, self.statsFrame.manaBar)
	XPerl_Unit_ManaBar_Update(self.statsFrame.manaBar, self.unit)
end
-- XPerl_Party_Pet_UpdateLevel
local function XPerl_Party_Pet_UpdateLevel(self)
	if (self.conf.level) then
		local Level = UnitLevel(self.unit)
		if Level == 0 then Level = "" end
		if (self.nameFrame.level) then
			self.nameFrame.level:SetText(Level)
			if (UnitCanAttack("player", self.unit)) then
				local color = GetQuestDifficultyColor(Level)
				self.nameFrame.level:SetVertexColor(color.r, color.g, color.b)
			else				
				self.nameFrame.level:SetVertexColor(1.0, 0.82, 0.0)
			end
		end
	end
end
-- XPerl_Party_Pet_Update_Control
local function XPerl_Party_Pet_Update_Control(self)
	if (self.unit and UnitIsVisible(self.unit) and UnitIsCharmed(self.unit) and not UnitUsingVehicle(self.ownerid)) then
		self.nameFrame.warningIcon:Show()
	else
		self.nameFrame.warningIcon:Hide()
	end
end
-- XPerl_Party_Pet_UpdateCombat
function XPerl_Party_Pet_UpdateCombat(self)
	if (self.unit and UnitIsVisible(self.unit) and UnitAffectingCombat(self.unit)) then
		self.nameFrame.level:Hide()
		self.nameFrame.combatIcon:Show()
	else
		self.nameFrame.combatIcon:Hide()
		if (petconf.level) then
			self.nameFrame.level:Show()
		end
	end
	XPerl_Party_Pet_Update_Control(self)
end

--------------------
-- Buff Functions --
--------------------
-- XPerl_Party_Pet_Buff_UpdateAll
function XPerl_Party_Pet_Buff_UpdateAll(self)
	if (self.unit) then
		if (petconf.buffs.enable) then
			if (UnitExists(self.unit)) then
				self.buffFrame:Show()
				if (XPerlDB) then
					if (not self.conf) then
						self.conf = conf.partypet
					end
	
					XPerl_Unit_UpdateBuffs(self, nil, nil, petconf.buffs.castble, petconf.debuffs.curable)
				end
			else
				self.buffFrame:Hide()
			end
		else
			self.buffFrame:Hide()
		end

		XPerl_CheckDebuffs(self, self.unit)
	end
end

-- XPerl_Party_Pet_UpdateDisplay
function XPerl_Party_Pet_UpdateDisplay(self)
	if (self.unit) then
		XPerl_Party_Pet_UpdateLevel(self)
		XPerl_Party_Pet_UpdateCombat(self)
		XPerl_Party_Pet_Buff_UpdateAll(self)
		XPerl_UpdateSpellRange(self)
	end
end
-- XPerl_Party_Pet_UpdateDisplayAll
function XPerl_Party_Pet_UpdateDisplayAll()
	for i,frame in pairs(PartyPetFrames) do
		if (frame:IsShown()) then
			XPerl_Party_Pet_UpdateDisplay(frame)
		end
	end
end

-- XPerl_Party_Pet_OnUpdate
function XPerl_Party_Pet_OnUpdate(self, elapsed)
	for i,f in pairs(PartyPetFrames) do
		if (f:IsShown()) then
			if (f.PlayerFlash) then
				XPerl_Party_Pet_CombatFlash(f, elapsed, false)
			end
			f.time = f.time + elapsed
			if (f.time > 0.2) then
				f.time = 0
				XPerl_UpdateSpellRange(f, nil, false)
			end
		end
	end
end

-------------------
-- Event Handler --
-------------------
-- XPerl_Party_Pet_OnEvent
function XPerl_Party_Pet_OnEvent(self, event, unitID, ...)
	--print (self:GetName().."-"..event.."-"..(unitID or "nil"))
	local func = XPerl_Party_Pet_Events[event]
	if (func) then
		if (strfind(event, "^UNIT_")) then
			local f = PartyPetFrames[unitID]
			if (f) then
				func(f, ...)
			end
		else
			func(unitID, ...)
		end
	else
		XPerl_Unit_OnEvent(self, event, unitID)
	end
end

-- PLAYER_ENTERING_WORLD
function XPerl_Party_Pet_Events:PLAYER_ENTERING_WORLD()
	XPerl_Party_Pet_UpdateGUIDs()
	XPerl_Party_Pet_UpdateDisplayAll()
end
XPerl_Party_Pet_Events.GROUP_ROSTER_UPDATE = XPerl_Party_Pet_Events.PLAYER_ENTERING_WORLD
XPerl_Party_Pet_Events.UNIT_PET = XPerl_Party_Pet_Events.PLAYER_ENTERING_WORLD
-- UNIT_FACTION, UNIT_FLAGS, UNIT_DYNAMIC_FLAGS
function XPerl_Party_Pet_Events:UNIT_FACTION()
	XPerl_Party_Pet_UpdateCombat(self)
	XPerl_Party_Pet_SetGray(self)
end
XPerl_Party_Pet_Events.UNIT_FLAGS = XPerl_Party_Pet_Events.UNIT_FACTION
XPerl_Party_Pet_Events.UNIT_DYNAMIC_FLAGS = XPerl_Party_Pet_Events.UNIT_FACTION
-- UNIT_NAME_UPDATE
function XPerl_Party_Pet_Events:UNIT_NAME_UPDATE()
	XPerl_Unit_UpdateName(self)
	XPerl_Party_Pet_UpdateGUIDs()
end
-- UNIT_LEVEL
function XPerl_Party_Pet_Events:UNIT_LEVEL()
	XPerl_Party_Pet_UpdateLevel(self)
end
-- UNIT_COMBAT
function XPerl_Party_Pet_Events:UNIT_COMBAT(...)
	local action, descriptor, damage, damageType = select(1, ...)
	
	if (action == "HEAL") then
		XPerl_Party_Pet_CombatFlash(self, 0, true, true)
	elseif (damage and damage > 0) then
		XPerl_Party_Pet_CombatFlash(self, 0, true)
	end
end
-- UNIT_AURA
function XPerl_Party_Pet_Events:UNIT_AURA()
	XPerl_Party_Pet_Buff_UpdateAll(self)
end

-- EnableDisable()
local function EnableDisable(self)
	if (petconf.enable) then
		RegisterUnitWatch(self)
	else
		UnregisterUnitWatch(self)
		self:Hide()
	end
end

-- XPerl_Party_Pet_Set_Bits1
function XPerl_Party_Pet_Set_Bits1(self)
	if (not self:GetParent()) then
		self:SetParent(_G["XPerl_party"..self:GetID()])
	end
	XPerl_Unit_Initialize(self, self.nameFrame.text, self.statsFrame.healthBar, self.statsFrame.manaBar, self.statsFrame.expectedHealth)

	if (petconf.name) then
		self.nameFrame:Show()
		self.nameFrame:SetHeight(20)
	else
		self.nameFrame:Hide()
		self.nameFrame:SetHeight(4)
	end

	if (petconf.mana) then
		self:SetHeight(50)
		self.statsFrame:SetHeight(33)
		self.statsFrame.manaBar:Show()
		if (petconf.percent) then
			self.statsFrame.manaBar.text:Show()
		else
			self.statsFrame.manaBar.text:Hide()
		end
	else
		self:SetHeight(40)
		self.statsFrame:SetHeight(23)
		self.statsFrame.manaBar:Hide()
	end

	if (petconf.level) then
		self.nameFrame.level:Show()
	else
		self.nameFrame.level:Hide()
	end

	self:SetScale(petconf.scale)

	XPerl_StatsFrameSetup(self)

	if (self:IsShown()) then
		XPerl_Party_Pet_UpdateDisplay(self)
	end

	local function SetAllBuffs(self, buffs)
		local prevAnchor
		if (pconf.flip) then
			prevAnchor = "TOPRIGHT"
		else
			prevAnchor = "TOPLEFT"
		end
		if (buffs) then
			local prev = self
			for k,v in pairs(buffs) do
				v:ClearAllPoints()
				if (pconf.flip) then
					v:SetPoint("TOPRIGHT", prev, prevAnchor, 0, 0)
				else
					v:SetPoint("TOPLEFT", prev, prevAnchor, 0, 0)
				end
				prev = v
				if (pconf.flip) then
					prevAnchor = "TOPLEFT"
				else
					prevAnchor = "TOPRIGHT"
				end
			end
		end
	end

	if (conf.highlight.enable and conf.highlight.HEAL) then
		self:RegisterEvent("UNIT_HEAL_PREDICTION")
	else
		self:UnregisterEvent("UNIT_HEAL_PREDICTION")
	end

	SetAllBuffs(self.buffFrame, self.buffFrame.debuff)
	SetAllBuffs(self.buffFrame, self.buffFrame.buff)
	local b = self.buffFrame.buff and self.buffFrame.buff[1]
	local d = self.buffFrame.debuff and self.buffFrame.debuff[1]
	if (b and d) then
		if (pconf.flip) then
			d:SetPoint("TOPRIGHT", b, "BOTTOMRIGHT", 0, 0)
		else
			d:SetPoint("TOPLEFT", b, "BOTTOMLEFT", 0, 0)
		end
	end

	XPerl_ProtectedCall(EnableDisable, self)
end

-- XPerl_Party_Pet_Set_Bits
function XPerl_Party_Pet_Set_Bits()
	if (XPerlDB) then
		conf = XPerlDB
		pconf = XPerlDB.party
		for k,v in pairs(PartyPetFrames) do
			v.conf = pconf
			XPerl_Party_Pet_Set_Bits1(v)
		end
	end

	local function RegisterEvents(self, enable, events)
		for k,v in pairs(events) do
			if (enable) then
				self:RegisterEvent(v)
			else
				self:UnregisterEvent(v)
			end
		end
	end
--[[
	RegisterEvents(XPerl_Party_Pet_EventFrame, petconf.mana, {"UNIT_POWER", "UNIT_MAXPOWER", "UNIT_DISPLAYPOWER"})
	RegisterEvents(XPerl_Party_Pet_EventFrame, petconf.name, {"UNIT_NAME_UPDATE"})
	RegisterEvents(XPerl_Party_Pet_EventFrame, petconf.name and petconf.level, {"UNIT_LEVEL"})
	]]
end
