-- X-Perl UnitFrames
-- Author: Zek <Boodhoof-EU>
-- License: GNU GPL v3, 29 June 2007 (see LICENSE.txt)

local XPerl_Target_Events = {}
local conf, tconf
XPerl_RequestConfig(function(new)
				conf = new
				tconf = conf.target
				if (XPerl_Target) then XPerl_Target.conf = conf.target end
				if (XPerl_Focus) then XPerl_Focus.conf = conf.focus end
				if (XPerl_TargetTarget) then XPerl_TargetTarget.conf = conf.targettarget end
				if (XPerl_FocusTarget) then XPerl_FocusTarget.conf = conf.focustarget end
				if (XPerl_PetTarget) then XPerl_PetTarget.conf = conf.pettarget end
			end, "$Revision: 652 $")

local UnitName = UnitName
local UnitIsConnected = UnitIsConnected
local GetQuestDifficultyColor = GetQuestDifficultyColor
local XPerl_Unit_Initialize = XPerl_Unit_Initialize
local buffSetup
local playerClass
local lastInspectPending = 0

----------------------
-- Loading Function --
----------------------
function XPerl_Target_OnLoad(self, unit)
	XPerl_SetChildMembers(self)
	
	self.hitIndicator.text:SetPoint("CENTER", self.portraitFrame, "CENTER", 0, 0)
	CombatFeedback_Initialize(self, self.hitIndicator.text, 30)

	self.unit = unit
	if (unit == "target") then
		XPerl_BlizzFrameDisable(TargetFrame)
		XPerl_BlizzFrameDisable(TargetofTargetFrame)
		self.statsFrame.focusTarget:SetVertexColor(0.7, 1, 1, 0.5)
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("PLAYER_FOCUS_CHANGED")
		self:RegisterUnitEvent("UNIT_COMBO_POINTS", "player")
		self:RegisterUnitEvent("PLAYER_FLAGS_CHANGED", "target")
		if (XPerl_Target_Events.INSPECT_READY) then
			self:RegisterEvent("INSPECT_READY")
		end
		self.combatMask = 0x00010000
	else
		XPerl_BlizzFrameDisable(FocusFrame)
		self:RegisterEvent("PLAYER_FOCUS_CHANGED")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterUnitEvent("PLAYER_FLAGS_CHANGED", "focus")
		self.combatMask = 0x00020000
	end
	
	local events = {"UNIT_AURA", "UNIT_COMBAT", "UNIT_LEVEL", "UNIT_FACTION", "UNIT_FLAGS", 
					"UNIT_DYNAMIC_FLAGS", "UNIT_THREAT_LIST_UPDATE", "UNIT_CLASSIFICATION_CHANGED", 
					"RAID_TARGET_UPDATE", "PARTY_LEADER_CHANGED", "PARTY_LOOT_METHOD_CHANGED", "GROUP_ROSTER_UPDATE"}
	for i,event in pairs(events) do
		if (strsub(event, 1, 5) == "UNIT_") then
			self:RegisterUnitEvent(event, unit)
		else
			self:RegisterEvent(event)
		end
	end
	
	self.nameFrame.cpMeter:SetMinMaxValues(0, 5)
	self.nameFrame.cpMeter:GetStatusBarTexture():SetHorizTile(false)
	self.nameFrame.cpMeter:GetStatusBarTexture():SetVertTile(false)

	local BuffOnUpdate, DebuffOnUpdate, BuffUpdateTooltip, DebuffUpdateTooltip
	BuffUpdateTooltip = XPerl_Unit_SetBuffTooltip
	DebuffUpdateTooltip = XPerl_Unit_SetDeBuffTooltip

	if (buffSetup) then
		self.buffSetup = buffSetup
	else
		self.buffSetup = {
			buffScripts = {
				OnEnter = XPerl_Unit_SetBuffTooltip,
				OnUpdate = BuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			debuffScripts = {
				OnEnter = XPerl_Unit_SetDeBuffTooltip,
				OnUpdate = DebuffOnUpdate,
				OnLeave = XPerl_PlayerTipHide,
			},
			updateTooltipBuff = BuffUpdateTooltip,
			updateTooltipDebuff = DebuffUpdateTooltip,
			debuffParent = true,
			debuffSizeMod = 0.2,
			debuffAnchor1 = function(self, b) b:SetPoint("TOPLEFT", 0, 0) end,
		}
		self.buffSetup.buffAnchor1 = self.buffSetup.debuffAnchor1
		buffSetup = self.buffSetup
	end

	XPerl_SecureUnitButton_OnLoad(self.nameFrame, unit, nil, TargetFrameDropDown, XPerl_ShowGenericMenu)
	XPerl_SecureUnitButton_OnLoad(self, unit, nil, TargetFrameDropDown, XPerl_ShowGenericMenu)

	self.perlBuffs, self.perlDebuffs, self.RangeUpdate = 0, 0, 0

	if (unit == "target" and XPerl_Target_AssistFrame) then
		-- Since target module is loaded after raid helper, we have to attach this manually
		-- because the target frame did not exist when this frame was created
		XPerl_Target_AssistFrame:SetParent(self)
		XPerl_Target_AssistFrame:ClearAllPoints()
		XPerl_Target_AssistFrame:SetPoint("TOPLEFT", self.portraitFrame, "TOPRIGHT", -2, -20)
		XPerl_Target_AssistFrame:Raise()
	end
	
	if (XPerlDB) then
		self.conf = XPerlDB[unit]
	end
	XPerl_RegisterOptionChanger(XPerl_Target_Set_Bits, self)
	
	XPerl_RegisterHighlight(self.highlight, 3)
	XPerl_Highlight:Register(XPerl_Target_HighlightCallback, self)
	XPerl_RegisterPerlFrames(self, {self.nameFrame, self.statsFrame, self.levelFrame, self.portraitFrame, self.classFrame, self.creatureTypeFrame, self.classTextFrame, self.cpFrame})

	self.FlashFrames = {self.portraitFrame, self.nameFrame, self.levelFrame, self.statsFrame, self.classTextFrame, self.classFrame, self.typeFrame}

	if (XPerl_Target and XPerl_Focus) then
		XPerl_Target_OnLoad = nil
	end
end

-- XPerl_Raid_HighlightCallback
function XPerl_Target_HighlightCallback(self, updateGUID)
	if (UnitGUID(self.unit) == updateGUID and UnitIsFriend("player", self.unit)) then
		XPerl_Highlight:SetHighlight(self, updateGUID)
	end
end
-- XPerl_Target_CombatFlash
local function XPerl_Target_CombatFlash(self, elapsed, argNew, argGreen)
	if (XPerl_CombatFlashSet (self, elapsed, argNew, argGreen)) then
		XPerl_CombatFlashSetFrames(self)
	end
end
-- XPerl_Target_SetGray
function XPerl_Target_SetGray(self)
	XPerl_Unit_SetGrey(self, self.statsFrame.healthBar)	
	XPerl_Unit_HealthBar_Update(self.statsFrame.healthBar, self.unit)
	XPerl_Unit_SetGrey(self, self.statsFrame.manaBar)
	XPerl_Unit_ManaBar_Update(self.statsFrame.manaBar, self.unit)
end
-- AdjustCreatureTypeFrame
local function AdjustCreatureTypeFrame(self)
	-- If it's too long, we anchor it to left side of portrait instead of right, to avoid it overlapping some buffs
	self.creatureTypeFrame:ClearAllPoints()
	if (self.creatureTypeFrame:GetWidth() > self.portraitFrame:GetWidth()) then
		self.creatureTypeFrame:SetPoint("TOPLEFT", self.portraitFrame, "BOTTOMLEFT", 0, 2)
	else
		self.creatureTypeFrame:SetPoint("TOPRIGHT", self.portraitFrame, "BOTTOMRIGHT", 0, 2)
	end
end
-- XPerl_Targets_BuffPositions
local function XPerl_Targets_BuffPositions(self)
	if (self.unit and UnitCanAttack("player", self.unit)) then
		XPerl_Unit_BuffPositions(self, self.buffFrame.debuff, self.buffFrame.buff, self.conf.debuffs.size, self.conf.buffs.size)
	else
		XPerl_Unit_BuffPositions(self, self.buffFrame.buff, self.buffFrame.debuff, self.conf.buffs.size, self.conf.debuffs.size)
	end
end
-- UnitFullName
local function UnitFullName(unit)
	local n,s = UnitName(unit)
	if (s and s ~= "") then
		return n.."-"..s
	end
	return n
end
-- RaidTargetUpdate
local function XPerl_Target_RaidIconUpdate(self)
	local raidIcon = self.nameFrame.raidIcon

	XPerl_Update_RaidIcon(raidIcon, self.unit)
	XPerl_Update_RaidIcon(self.eliteFrame.raidIcon, self.unit)

	raidIcon:ClearAllPoints()
	if (self.conf.raidIconAlternate) then
		raidIcon:SetHeight(20)
		raidIcon:SetWidth(20)
		raidIcon:SetPoint("CENTER", self.nameFrame, "TOPRIGHT", -5, -4)
		if self.eliteFrame:IsShown() then 
			self.nameFrame.raidIcon:Hide()
		end
	else
		raidIcon:SetHeight(32)
		raidIcon:SetWidth(32)
		raidIcon:SetPoint("CENTER", self.nameFrame, "CENTER", 0, 0)
		self.eliteFrame.raidIcon:Hide()
	end
end
-- XPerl_Target_UpdateLevel
function XPerl_Target_UpdateLevel(self)
	if (self.conf.level) then
		local targetLevel
		local isBattlePet = UnitIsWildBattlePet(self.unit) or UnitIsBattlePetCompanion(self.unit)
		self.levelFrame:SetWidth(27)
		if (isBattlePet) then
			targetLevel = UnitBattlePetLevel(self.unit)
		else
			targetLevel = UnitLevel(self.unit)
		end
		if (targetLevel > 0) then
			if (UnitCanAttack("player", self.unit) and not (isBattlePet)) then
				local color = GetQuestDifficultyColor(targetLevel)
				self.levelFrame.text:SetVertexColor(color.r, color.g, color.b)
			else				
				self.levelFrame.text:SetVertexColor(1.0, 0.82, 0.0)
			end
			if (not self.conf.elite and (UnitClassification(self.unit) == "elite" or UnitClassification(self.unit) == "worldboss")) then
				self.levelFrame.text:SetFormattedText("%d+", targetLevel)
				self.levelFrame:SetWidth(33)
			else
				self.levelFrame.text:SetText(targetLevel)
			end
			self.levelFrame.text:Show()
			self.levelFrame.skull:Hide()
		else
			self.levelFrame.text:Hide()
			self.levelFrame.skull:Show()
		end
	end
end
-- XPerl_Target_UpdateTalents
local XPerl_Target_UpdateTalents
do
	local function ShowSpec(self, spec)
		if (self.conf.talentsAsText and type(spec) == "string") then
			self.creatureTypeFrame.text:SetText(spec)
		end
		self.creatureTypeFrame:SetWidth(self.creatureTypeFrame.text:GetStringWidth() + 10)
		self.creatureTypeFrame:Show()

		AdjustCreatureTypeFrame(self)
		XPerl_Targets_BuffPositions(self)
	end

	local inspectReady
	local talentCache = setmetatable({}, {__mode = "kv"})
	hooksecurefunc("NotifyInspect",
		function(unit)
			if (UnitIsUnit("player", unit) or UnitInVehicle(unit) or not (UnitExists(unit) and CanInspect(unit) and UnitIsVisible(unit) and UnitIsConnected(unit))) then
				return
			end
			lastInspectPending = lastInspectPending + 1
			if (lastInspectPending > 1) then
				lastInspectInvalid = true
			end
		end
	)

	-- INSPECT_READY
	function XPerl_Target_Events:INSPECT_READY(guid)
		if (UnitGUID(self.unit) == guid) then
			inspectReady = true
			XPerl_Target_UpdateTalents(self, guid)
		end
	end

	function XPerl_Target_UpdateTalents(self, guid)
		if (self.conf.showTalents and self == XPerl_Target and not self.creatureTypeFrame:IsShown()) then
			local unit = self.unit
			if (UnitIsVisible(unit) and UnitExists(unit) and UnitIsPlayer(unit) and UnitLevel(unit) > 10) then
				local name = UnitFullName(unit)
				if (not name) then
					return
				else
					local specName
					local cached = talentCache[name]
					if (cached) then
						specName, checkedTime = unpack(cached)
						if GetTime() > checkedTime + 15 then 
							talentCache[name] = nil
							specName = nil
							cached = nil
						end
					elseif (inspectReady and guid == UnitGUID(unit)) then
						local spec = GetInspectSpecialization(inspectedUnit)
						if GetSpecializationRoleByID(spec) then
							specName = select(2,GetSpecializationInfoByID(spec))
						end
						inspectReady = nil
						lastInspectPending = 0
					end

					if (specName) then
						if (not cached) then
							talentCache[name] = {specName, GetTime()}
						end
						ShowSpec(self, specName)
						return
					end

					if (not cached) then
						if (lastInspectPending == 0 and UnitExists(unit) and UnitIsVisible(unit)) then
							if (not UnitIsUnit("player", unit)) then
								inspectReady = nil
								lastInspectInvalid = nil
								lastInspectPending = 0
								inspectedUnit = unit
								NotifyInspect(unit)
							end
						end
					end
				end
			end
		end
	end
end
-- XPerl_Target_UpdateClassification
local function XPerl_Target_UpdateClassification(self)
	local targetclassification = UnitClassification(self.unit)
	local mobType, eliteGfx

	if (targetclassification == "normal" and UnitPlayerControlled(self.unit)) then
		if (not UnitIsPlayer(self.unit)) then
			mobType = XPERL_TYPE_PET
			self.classTextFrame.text:SetTextColor(1.0, 1.0, 1.0)
			self.classFrame:Hide()
		end
	elseif (self.conf.level or (targetclassification ~= "normal" and targetclassification ~= "trivial")) then
		if (self.conf.eliteGfx) then
			eliteGfx = true
			self.eliteFrame.tex:SetTexture("Interface\\Addons\\XPerl\\Images\\XPerl_Elite")
			if (targetclassification == "worldboss" or targetclassification == "elite") then
				self.eliteFrame.tex:SetVertexColor(1.0, 0.8, 0.0, 1.0)
			elseif (targetclassification == "rareelite") then
				self.eliteFrame.tex:SetVertexColor(1.0, 1.0, 1.0, 1.0)
			elseif (targetclassification == "rare") then
				self.eliteFrame.tex:SetTexture("Interface\\Addons\\XPerl\\Images\\XPerl_Rare")
				self.eliteFrame.tex:SetVertexColor(1.0, 1.0, 1.0, 1.0)
			else
				eliteGfx = nil
			end
		else
			if (targetclassification == "worldboss") then
				mobType = XPERL_TYPE_BOSS
				self.classTextFrame.text:SetTextColor(1.0, 0.5, 0.5)
			elseif (targetclassification == "rareelite") then
				mobType = XPERL_TYPE_RAREPLUS
				self.classTextFrame.text:SetTextColor(0.8, 0.8, 0.8)
			elseif (targetclassification == "elite") then
				mobType = XPERL_TYPE_ELITE
				self.classTextFrame.text:SetTextColor(1.0, 1.0, 0.5)
			elseif (targetclassification == "rare") then
				mobType = XPERL_TYPE_RARE
				self.classTextFrame.text:SetTextColor(0.8, 0.8, 0.8)
			end
		end
		self.classFrame:Hide()
	end

	if (mobType) then
		self.classTextFrame:Show()
		self.classTextFrame.text:SetText(mobType)
		self.classTextFrame:SetWidth(self.classTextFrame.text:GetStringWidth() + 10)
	else
		self.classTextFrame:Hide()
	end

	if (eliteGfx) then
		self.eliteFrame:Show()
		if (XPerl_Target_AssistFrame and XPerl_Target_AssistFrame:IsShown()) then
			XPerl_Target_AssistFrame:ClearAllPoints()
			XPerl_Target_AssistFrame:SetPoint("BOTTOMRIGHT", self.portraitFrame, "BOTTOMRIGHT", 0, 0)
			XPerl_Target_AssistFrame:SetFrameLevel(self.portraitFrame:GetFrameLevel() + 2)
			self.eliteFrame.assistOutOfPlace = true
		end
		if (self.conf.level) then
			self.levelFrame:ClearAllPoints()
			self.levelFrame:SetPoint("TOPRIGHT", self.portraitFrame, "TOPRIGHT", 0, 0)
			self.levelFrame:SetFrameLevel(self.portraitFrame:GetFrameLevel() + 2)
			self.levelFrame.outOfPlace = true
		end
	else
		self.eliteFrame:Hide()
		if (self.eliteFrame.assistOutOfPlace) then
			self.eliteFrame.assistOutOfPlace = nil
			XPerl_Target_AssistFrame:ClearAllPoints()
			XPerl_Target_AssistFrame:SetPoint("TOPLEFT", self.portraitFrame, "TOPRIGHT", -2, -20)
			XPerl_Target_AssistFrame:SetFrameLevel(self.portraitFrame:GetFrameLevel())
		end
		if (self.levelFrame.outOfPlace) then
			self.levelFrame.outOfPlace = nil
			self.levelFrame:ClearAllPoints()
			self.levelFrame:SetPoint("TOPLEFT", self.portraitFrame, "TOPRIGHT", -2, 0)
			self.levelFrame:SetFrameLevel(self.portraitFrame:GetFrameLevel())
		end
	end
end
-- XPerl_Target_UpdateType
local function XPerl_Target_UpdateType(self)
	local targettype = UnitCreatureType(self.unit)
	
	self.classFrame:Hide()
	self.creatureTypeFrame:Hide()
	
	if (targettype == XPERL_TYPE_NOT_SPECIFIED or targettype == "") then
		targettype = nil
	end
	if (not UnitIsPlayer(self.unit) and (self.conf.mobType and targettype)) then
		self.creatureTypeFrame:Show()
		self.creatureTypeFrame.text:SetText(targettype)
		self.creatureTypeFrame.text:SetTextColor(1, 1, 1)
		self.creatureTypeFrame:SetWidth(self.creatureTypeFrame.text:GetStringWidth() + 10)
		AdjustCreatureTypeFrame(self)
	end
	
	if (self.conf.classIcon and (UnitIsPlayer(self.unit) or UnitClassification(self.unit) == "normal")) then
		local LocalClass, PlayerClass = UnitClassBase(self.unit)
		if (self.conf.classText) then
			self.classTextFrame:Show()
			self.classTextFrame.text:SetText(LocalClass)
			self.classTextFrame.text:SetTextColor(1, 1, 1)
			self.classTextFrame:SetWidth(self.classTextFrame.text:GetStringWidth() + 10)
		else
			if (UnitIsPlayer(self.unit) or not UnitPlayerControlled(self.unit)) then
				local l, r, t, b = XPerl_ClassPos(PlayerClass)
				self.classFrame.tex:SetTexCoord(l, r, t, b)
				self.classFrame:Show()
			end
		end
	end
	XPerl_Target_UpdateTalents(self)
end
local function XPerl_Target_CheckBattlePet(self)
	if ( UnitIsWildBattlePet(self.unit) or UnitIsBattlePetCompanion(self.unit) ) then
		local petType = UnitBattlePetType(self.unit)
		self.nameFrame.petBattleIcon:SetTexture("Interface\\TargetingFrame\\PetBadge-"..PET_TYPE_SUFFIX[petType])
		self.nameFrame.petBattleIcon:Show()
	else
		self.nameFrame.petBattleIcon:Hide()
	end
end
-- XPerl_Target_Update_Combat
function XPerl_Target_Update_Combat(self)
	if (UnitAffectingCombat(self.unit)) then
		self.nameFrame.combatIcon:Show()
	else
		self.nameFrame.combatIcon:Hide()
	end
end
-- XPerl_Target_UpdateLeader
local function XPerl_Target_UpdateLeader(self)
	-- Leader
	if UnitLeadsAnyGroup(self.unit) then
		self.nameFrame.leaderIcon:Show()
	else
		self.nameFrame.leaderIcon:Hide()
	end
	-- Loot Master
	local isMasterLooter
	local method, mlPartyID, mlRaidID  = GetLootMethod()
	if (method == "master" and (mlPartyID or mlRaidID)) then
		if (mlRaidID) then 
			if (mlRaidID == UnitInRaid(self.unit)) then 
				isMasterLooter = true 
			end
		elseif (mlPartyID == 0 and UnitIsUnit("player", self.unit)) then
			isMasterLooter = true
		elseif (mlPartyID >= 1 and mlPartyID <= 4) then
			isMasterLooter = UnitIsUnit(self.unit, "party"..mlPartyID)
		end
	end

	if (isMasterLooter) then
		self.nameFrame.masterIcon:Show()
	else
		self.nameFrame.masterIcon:Hide()
	end
end
function XPerl_Target_UpdatePVP(self)
	local factionGroup = UnitFactionGroup(self.unit)
	if (self.conf.pvpIcon and UnitIsPVPFreeForAll(self.unit)) then
		self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-FFA")
		self.nameFrame.pvpIcon:Show()
	elseif (self.conf.pvpIcon and factionGroup and factionGroup ~= "Neutral" and UnitIsPVP(self.unit)) then
		self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..factionGroup)
		self.nameFrame.pvpIcon:Show()
	else
		self.nameFrame.pvpIcon:Hide()
	end

	if (UnitIsVisible(self.unit) and UnitIsCharmed(self.unit)) then
		self.nameFrame.warningIcon:Show()
	else
		self.nameFrame.warningIcon:Hide()
	end
end
--------------------
-- Buff Functions --
--------------------
-- XPerl_Target_CheckDebuffs
local function XPerl_Target_CheckDebuffs(self)
	if (self.conf.highlightDebuffs and self.conf.highlightDebuffs.enable) then
		if (self.conf.highlightDebuffs.who == 1 or
			(self.conf.highlightDebuffs.who == 2 and UnitCanAssist("player", self.unit)) or
			(self.conf.highlightDebuffs.who == 3 and not UnitCanAssist("player", self.unit))) then

			XPerl_CheckDebuffs(self, self.unit)
		else
			XPerl_CheckDebuffs(self, self.unit, true)
		end
	end
end
-- XPerl_Targets_BuffUpdate
function XPerl_Targets_BuffUpdate(self)
	if (UnitExists(self.unit)) then
		XPerl_Target_CheckDebuffs(self)
		if (not self.conf.buffs.enable and not self.conf.debuffs.enable) then
			self.buffFrame:Hide()
			self.debuffFrame:Hide()
		else
			XPerl_Unit_UpdateBuffs(self, nil, nil, self.conf.buffs.castable, self.conf.debuffs.curable)
			XPerl_Targets_BuffPositions(self)
		end
	end
end
------------------
-- Combo Points --
------------------
-- GetComboColour
local function GetComboColour(num)
	if	   (num == 5) then	return 1.0, 0.0, 0.0
	elseif (num == 4) then	return 1.0, 0.5, 0.0
	elseif (num == 3) then	return 1.0, 1.0, 0.0
	elseif (num == 2) then	return 0.5, 1.0, 0.0
	elseif (num == 1) then	return 0.0, 1.0, 0.0
	end
end
local function XPerl_Target_UpdateCombo(self)
	local combopoints = GetComboPoints(UnitHasVehicleUI("player") and "vehicle" or "player", self.unit)
	local r, g, b = GetComboColour(combopoints)
	if (tconf.combo.enable) then
		self.cpFrame:Hide()
		self.nameFrame.cpMeter:SetValue(combopoints)
		self.nameFrame.cpMeter:Show()
		if (r) then
			self.nameFrame.cpMeter:SetStatusBarColor(r, g, b, 0.7)
		else
			self.nameFrame.cpMeter:Hide()
		end
	else
		self.nameFrame.cpMeter:Hide()
		if (not tconf.combo.blizzard) then
			self.cpFrame:Show()
			self.cpFrame.text:SetText(combopoints)
			if (r) then
				self.cpFrame.text:SetTextColor(r, g, b)
			else
				self.cpFrame:Hide()
			end
		end
	end
end
-- XPerl_Target_Update_Range
function XPerl_Target_Update_Range(self)
	if (not self.conf.range30yard or CheckInteractDistance(self.unit, 1) or not UnitIsConnected(self.unit)) then
		self.nameFrame.rangeIcon:Hide()
	else
		self.nameFrame.rangeIcon:Show()
		self.nameFrame.rangeIcon:SetAlpha(1)
	end
end

-- XPerl_Target_UpdateDisplay
function XPerl_Target_UpdateDisplay(self)
	if (UnitExists(self.unit)) then
		XPerl_NoFadeBars(true)
		
		XPerl_Unit_UpdateName(self)
		XPerl_Target_UpdateLevel(self)
		XPerl_Target_UpdateClassification(self)
		XPerl_Target_UpdateType(self)
		XPerl_Target_CheckBattlePet(self)
		XPerl_Target_RaidIconUpdate(self)
		XPerl_Target_SetGray(self)
		XPerl_Unit_HealPredictionBars_Update(self)
		XPerl_Unit_ThreatStatus(self, self.unit == "target" and "player" or nil, true)

		if (self.conf.defer) then
			self.portraitFrame.portrait:Hide()
			self.portraitFrame.portrait3D:Hide()
			self.nameFrame.leaderIcon:Hide()
			self.nameFrame.masterIcon:Hide()
			self.cpFrame:Hide()
			self.nameFrame.cpMeter:Hide()
			self.deferring = true
			self.RangeUpdate = -0.3
		else
			XPerl_Target_UpdateLeader(self)
			XPerl_Target_Update_Combat(self)
			XPerl_Target_UpdateCombo(self)
			XPerl_Unit_UpdatePortrait(self)
		end
		
		XPerl_NoFadeBars()
		XPerl_Targets_BuffUpdate(self)
		XPerl_Target_UpdatePVP(self)

		XPerl_Highlight:SetHighlight(self, UnitGUID(self.unit))
		XPerl_UpdateSpellRange(self, self.unit)
	end
end

-------------------------
-- The Update Function --
-------------------------
function XPerl_Target_OnUpdate(self, elapsed)
	if (self.PlayerFlash) then
		XPerl_Target_CombatFlash(self, elapsed, false)
	end

	self.RangeUpdate = self.RangeUpdate + elapsed
	if (self.RangeUpdate > 0.2) then
		self.RangeUpdate = 0
		XPerl_Target_Update_Range(self)
		XPerl_UpdateSpellRange(self)

		if (self.deferring) then
			self.deferring = nil
			XPerl_Target_UpdateLeader(self)
			XPerl_Target_Update_Combat(self)
			XPerl_Target_UpdateCombo(self)
			XPerl_Unit_UpdatePortrait(self)
		end
	end
	
	CombatFeedback_OnUpdate(self, elapsed)
end

-------------------
-- Event Handler --
-------------------
function XPerl_Target_OnEvent(self, event, unitID, ...)
	--print (self:GetName().."-"..event.."-"..(unitID or "nil"))
	local func = XPerl_Target_Events[event]
	if (func) then
		func(self, unitID, ...)
	else
		XPerl_Unit_OnEvent(self, event, unitID)
	end
end

-- PLAYER_ENTERING_WORLD
function XPerl_Target_Events:PLAYER_ENTERING_WORLD()
	if (UnitExists("focus")) then
		self.PlayerFlash = 0
		XPerl_CombatFlashSetFrames(self)
		XPerl_Target_UpdateDisplay(self)
	end
end
-- UNIT_LEVEL
function XPerl_Target_Events:UNIT_LEVEL()
	XPerl_Target_UpdateLevel(self)
	XPerl_Target_UpdateClassification(self)
end
-- RAID_TARGET_UPDATE
function XPerl_Target_Events:RAID_TARGET_UPDATE()
	XPerl_Target_RaidIconUpdate(XPerl_Target)
	XPerl_Target_RaidIconUpdate(XPerl_Focus)
end
-- UNIT_COMBO_POINTS
function XPerl_Target_Events:UNIT_COMBO_POINTS(unit)
	if (unit == "player") or (unit == "vehicle") then
		XPerl_Target_UpdateCombo(self)
	end
end
-- UNIT_CLASSIFICATION_CHANGED
function XPerl_Target_Events:UNIT_CLASSIFICATION_CHANGED()
	XPerl_Target_UpdateClassification(self)
end
-- UNIT_FACTION
function XPerl_Target_Events:UNIT_FACTION()
	XPerl_Target_UpdatePVP(self)
	XPerl_Targets_BuffPositions(self)
end
-- UNIT_DYNAMIC_FLAGS
function XPerl_Target_Events:UNIT_FLAGS()
	XPerl_Target_UpdatePVP(self)
	XPerl_Target_Update_Combat(self)
	XPerl_Target_SetGray(self)
end
XPerl_Target_Events.UNIT_DYNAMIC_FLAGS = XPerl_Target_Events.UNIT_FLAGS
XPerl_Target_Events.PLAYER_FLAGS_CHANGED = XPerl_Target_Events.UNIT_FLAGS
-- PARTY_LOOT_METHOD_CHANGED, PARTY_LEADER_CHANGED, GROUP_ROSTER_UPDATE
function XPerl_Target_Events:PARTY_LOOT_METHOD_CHANGED()
	XPerl_Target_UpdateLeader(self)
end
XPerl_Target_Events.GROUP_ROSTER_UPDATE = XPerl_Target_Events.PARTY_LOOT_METHOD_CHANGED
XPerl_Target_Events.PARTY_LEADER_CHANGED  = XPerl_Target_Events.PARTY_LOOT_METHOD_CHANGED
-- PLAYER_TARGET_CHANGED
function XPerl_Target_Events:PLAYER_TARGET_CHANGED()
	if (self == XPerl_Target) then
		if (self.conf.sound and UnitExists("target")) then
			if (UnitIsEnemy("target", "player")) then
				PlaySound("igCreatureAggroSelect")
			elseif (UnitIsFriend("player", "target")) then
				PlaySound("igCharacterNPCSelect")
			else
				PlaySound("igCreatureNeutralSelect")
			end
		end

		if (UnitIsUnit("target", "focus")) then
			XPerl_Target.statsFrame.focusTarget:Show()
		else
			XPerl_Target.statsFrame.focusTarget:Hide()
		end
	end

	self.PlayerFlash = 0
	XPerl_CombatFlashSetFrames(self)
	XPerl_Target_UpdateDisplay(self)
end
XPerl_Target_Events.PLAYER_FOCUS_CHANGED = XPerl_Target_Events.PLAYER_TARGET_CHANGED
-- UNIT_THREAT_LIST_UPDATE
function XPerl_Target_Events:UNIT_THREAT_LIST_UPDATE(unit)
	if (UnitCanAttack("player", self.unit or "target")) then
		XPerl_Unit_ThreatStatus(self, self.unit == "target" and "player" or nil)
	else
		XPerl_Unit_ThreatStatus(self)
	end
end
-- UNIT_AURA
function XPerl_Target_Events:UNIT_AURA()
	XPerl_Targets_BuffUpdate(self)
end
-- UNIT_COMBAT
function XPerl_Target_Events:UNIT_COMBAT(...)
	local unitID, action, descriptor, damage, damageType = select(1, ...)

	if (unitID == self.unit) then
		if (not self.conf.ownDamageOnly) then
			if (self.conf.hitIndicator and self.conf.portrait) then
				CombatFeedback_OnCombatEvent(self, action, descriptor, damage, damageType)
			end
		end

		if (action == "HEAL") then
			XPerl_Target_CombatFlash(self, 0, true, true)
		elseif (damage and damage > 0) then
			XPerl_Target_CombatFlash(self, 0, true)
		end
	end
end
----------------------------
-- ONLY_OWN_HIT_INDICATOR --
----------------------------
local bit_band = bit.band
local amountIndex = {
	SWING_DAMAGE = 1,
	RANGE_DAMAGE = 4,
	SPELL_DAMAGE = 4,
	SPELL_PERIODIC_DAMAGE = 4,
	DAMAGE_SHIELD = 4,
	ENVIRONMENTAL_DAMAGE = 2,
}
local missIndex = {
	SWING_MISSED = 1,
	RANGE_MISSED = 4,
	SPELL_MISSED = 4,
	SPELL_PERIODIC_MISSED = 4,
}
-- DoEvent
local function DoEvent(self, timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	local feedbackText = self.feedbackText
	local fontHeight = self.feedbackFontHeight
	local text
	local r = 1
	local g = 1
	local b = 1

	if (event == "SWING_DAMAGE" or event == "RANGE_DAMAGE" or event == "SPELL_DAMAGE" or event == "DAMAGE_SHIELD" or event == "ENVIRONMENTAL_DAMAGE" or event == "SPELL_PERIODIC_DAMAGE") then
		local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(amountIndex[event], ...)
		if (amount and amount ~= 0) then
			if (critical or crushing) then
				fontHeight = fontHeight * 1.5
			elseif (glancing) then
				fontHeight = fontHeight * 0.75
			end
			if (event ~= "SWING_DAMAGE" and event ~= "RANGE_DAMAGE") then
				b = 0
			end
			text = amount
		end

	elseif (event == "SWING_MISSED" or event == "RANGE_MISSED" or event == "SPELL_MISSED" or event == "SPELL_PERIODIC_MISSED") then
		local missType = select(missIndex[event], ...)
		if (missType) then
			text = CombatFeedbackText[missType]
		end

	elseif (event == "SPELL_HEAL" or event == "SPELL_PERIODIC_HEAL") then
		local amount, overhealing, absorbed, critical = select(4, ...)
		if (amount and amount ~= 0) then
			if (critical) then
				fontHeight = fontHeight * 1.5
			end
			r = 0
			g = 1
			b = 0
			text = amount
		end
	end

	if (text) then
		self.feedbackStartTime = GetTime()

		feedbackText:SetTextHeight(fontHeight)
		feedbackText:SetText(text)
		feedbackText:SetTextColor(r, g, b)
		feedbackText:SetAlpha(0)
		feedbackText:Show()
	end
end
-- COMBAT_LOG_EVENT_UNFILTERED
function XPerl_Target_Events:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, ...)
	if (self.conf.hitIndicator and self.conf.portrait) then
		if (bit_band(dstFlags, self.combatMask) ~= 0 and bit_band(srcFlags, 0x00000001) ~= 0) then
			DoEvent(self, timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
		end
	end
end

-- XPerl_Target_SetWidth
function XPerl_Target_SetWidth(self)

	self.conf.size.width = max(0, self.conf.size.width or 0)
	local w = 128 + ((self.conf.portrait or 0) * 60) + ((self.conf.percent or 0) * 32) + self.conf.size.width

	if (not InCombatLockdown()) then
		self:SetWidth(w)
	end

	if (self.conf.percent) then
		if (not InCombatLockdown()) then
			self.nameFrame:SetWidth(160 + self.conf.size.width)
			self.statsFrame:SetWidth(160 + self.conf.size.width)
		end
		self.statsFrame.healthBar.percent:Show()
		self.statsFrame.manaBar.percent:Show()
	else
		if (not InCombatLockdown()) then
			self.nameFrame:SetWidth(128 + self.conf.size.width)
			self.statsFrame:SetWidth(128 + self.conf.size.width)
		end
		self.statsFrame.healthBar.percent:Hide()
		self.statsFrame.manaBar.percent:Hide()
	end

	self.conf.scale = self.conf.scale or 0.8
	if (not InCombatLockdown()) then
		self:SetScale(self.conf.scale)
	end

	XPerl_SavePosition(self, true)
end

-- XPerl_Target_Set_Bits
function XPerl_Target_Set_Bits(self)
	XPerl_Unit_Initialize(self, self.nameFrame.text, self.statsFrame.healthBar, self.statsFrame.manaBar, self.statsFrame.expectedHealth)
	if (XPerl_ArcaneBar_RegisterFrame) then
		XPerl_ArcaneBar_RegisterFrame(self.nameFrame, self.unit)
	end
	
	playerClass = select(2,UnitClass("player"))
	
	if (self.conf.portrait) then
		self.portraitFrame:Show()
		self.portraitFrame:SetWidth(60)
	else
		self.portraitFrame:Hide()
		self.portraitFrame:SetWidth(3)
	end
	
	self.eliteFrame:SetFrameLevel(self.portraitFrame:GetFrameLevel() + 3)
	if (self == XPerl_Target) then
		XPerl_Target_Set_BlizzCPFrame(self)
	end
	
	if (self.conf.ownDamageOnly and (self.conf.hitIndicator and self.conf.portrait)) then
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	else
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
	
	if (self.conf.level) then
		self.levelFrame:Show()
	else
		self.levelFrame:Hide()
	end

	if (self.conf.values) then
		self.statsFrame.healthBar.text:Show()
		self.statsFrame.manaBar.text:Show()
	else
		self.statsFrame.healthBar.text:Hide()
		self.statsFrame.manaBar.text:Hide()
	end
	
	XPerl_Target_SetWidth(self)

	self.highlight:SetPoint("BOTTOMRIGHT", self.portraitFrame, "BOTTOMRIGHT", 26, -1)

	self.buffOptMix = nil
	self.buffFrame:ClearAllPoints()
	if (self.conf.buffs.above) then
		self.buffFrame:SetPoint("BOTTOMLEFT", self, "TOPLEFT", 2, 0)
	else
		self.buffFrame:SetPoint("TOPLEFT", self.statsFrame, "BOTTOMLEFT", 2, 0)
	end
	self.conf.buffs.size = tonumber(self.conf.buffs.size) or 20
	XPerl_SetBuffSize(self)
	XPerl_StatsFrameSetup(self)
	
	if (not InCombatLockdown()) then
		if (self.conf.enable) then
			RegisterUnitWatch(self)
		else
			self:Hide()
			UnregisterUnitWatch(self)
		end
	end

	if (self:IsShown()) then
		XPerl_Target_UpdateDisplay(self)
	end
end

-- Using the Blizzard Combo Point frame, but we move the buttons around a little
function XPerl_Target_Set_BlizzCPFrame(self)
	ComboFrame:ClearAllPoints()
	ComboPoint1:ClearAllPoints()
	ComboPoint2:ClearAllPoints()
	ComboPoint3:ClearAllPoints()
	ComboPoint4:ClearAllPoints()
	ComboPoint5:ClearAllPoints()

	if (tconf.combo.blizzard) then
		-- This setup will put the buttons along the top of the portrait frame

		ComboFrame:SetScale(self.conf.scale)

		if (tconf.combo.pos == "top") then
			ComboFrame:SetPoint("TOPLEFT", self.portraitFrame, "TOPLEFT", -1, 4)
			ComboPoint1:SetPoint("TOPLEFT", 0, 0)
			ComboPoint2:SetPoint("LEFT", ComboPoint1, "RIGHT", 0, 1)
			ComboPoint3:SetPoint("LEFT", ComboPoint2, "RIGHT", 0, 1)
			ComboPoint4:SetPoint("LEFT", ComboPoint3, "RIGHT", 0, -1)
			ComboPoint5:SetPoint("LEFT", ComboPoint4, "RIGHT", 0, -4)

		elseif (tconf.combo.pos == "bottom") then
			ComboFrame:SetPoint("BOTTOMLEFT", self.portraitFrame, "BOTTOMLEFT", -1, -4)
			ComboPoint1:SetPoint("BOTTOMLEFT", 0, 0)
			ComboPoint2:SetPoint("LEFT", ComboPoint1, "RIGHT", 0, -1)
			ComboPoint3:SetPoint("LEFT", ComboPoint2, "RIGHT", 0, -1)
			ComboPoint4:SetPoint("LEFT", ComboPoint3, "RIGHT", 0, 1)
			ComboPoint5:SetPoint("LEFT", ComboPoint4, "RIGHT", 0, -1)

		elseif (tconf.combo.pos == "left") then
			ComboFrame:SetPoint("BOTTOMLEFT", self.portraitFrame, "BOTTOMLEFT", -1, 0)
			ComboPoint1:SetPoint("BOTTOMLEFT", 0, 0)
			ComboPoint2:SetPoint("BOTTOM", ComboPoint1, "TOP", -1, 0)
			ComboPoint3:SetPoint("BOTTOM", ComboPoint2, "TOP", -1, 0)
			ComboPoint4:SetPoint("BOTTOM", ComboPoint3, "TOP", 1, 0)
			ComboPoint5:SetPoint("BOTTOM", ComboPoint4, "TOP", 3, -5)

		elseif (tconf.combo.pos == "right") then
			ComboFrame:SetPoint("BOTTOMRIGHT", self.portraitFrame, "BOTTOMRIGHT", 2, 0)
			ComboPoint1:SetPoint("BOTTOMRIGHT", 0, 0)
			ComboPoint2:SetPoint("BOTTOM", ComboPoint1, "TOP", 1, 0)
			ComboPoint3:SetPoint("BOTTOM", ComboPoint2, "TOP", 1, 0)
			ComboPoint4:SetPoint("BOTTOM", ComboPoint3, "TOP", -1, 0)
			ComboPoint5:SetPoint("BOTTOM", ComboPoint4, "TOP", 0, -5)

		end

		function ComboFrame_OnEvent(self, event, ...)
			if (event == "PLAYER_TARGET_CHANGED") then
				ComboFrame_Update()
			elseif (event == "UNIT_COMBO_POINTS") then
				local unit = ...
				if (unit == "player" or unit == "vehicle") then
					ComboFrame_Update()
				end
			end
		end

		ComboFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
		ComboFrame:RegisterUnitEvent("UNIT_COMBO_POINTS", "player")
		ComboFrame:SetScript("OnEvent", ComboFrame_OnEvent)

		function ComboFrame_Update()
			local comboPoints = GetComboPoints(UnitHasVehicleUI("player") and "vehicle" or "player")
			local comboPoint, comboPointHighlight, comboPointShine
			if ( comboPoints > 0 ) then
				if ( not ComboFrame:IsShown() ) then
					ComboFrame:Show()
					UIFrameFadeIn(ComboFrame, COMBOFRAME_FADE_IN)
				end

				local fadeInfo = {}
				for i=1, MAX_COMBO_POINTS do
					comboPointHighlight = _G["ComboPoint"..i.."Highlight"]
					comboPointShine = _G["ComboPoint"..i.."Shine"]
					if ( i <= comboPoints ) then
						if ( i > COMBO_FRAME_LAST_NUM_POINTS ) then
							-- Fade in the highlight and set a function that triggers when it is done fading
							fadeInfo.mode = "IN"
							fadeInfo.timeToFade = COMBOFRAME_HIGHLIGHT_FADE_IN
							fadeInfo.finishedFunc = ComboPointShineFadeIn
							fadeInfo.finishedArg1 = comboPointShine
							UIFrameFade(comboPointHighlight, fadeInfo)
						end
					else
						comboPointHighlight:SetAlpha(0)
						comboPointShine:SetAlpha(0)
					end
				end
			else
				ComboPoint1Highlight:SetAlpha(0)
				ComboPoint1Shine:SetAlpha(0)
				ComboFrame:Hide()
			end
			COMBO_FRAME_LAST_NUM_POINTS = comboPoints
		end
	else
		ComboFrame:Hide()
		ComboFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
		ComboFrame:UnregisterEvent("UNIT_COMBO_POINTS")
	end
end
